#include "gsmhacking.h"

/* make a 52-bone SWT model */
int make_swt(struct gsm *gsm) {
  int type, n;
 
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  n = gsm->type & GSM_TYPE_ANIMATION;
  type = gsm->type % GSM_TYPE_ANIMATION_LIMITED_BONES;

  switch (type) {
    case GSM_TYPE_SWT:
      error_dbg("GSM is already a SWT model\n");
      return 0;
    case GSM_TYPE_KID:
    SAY("converting KID %s to SWT\n", (n) ? "animation" : "model");
      if (n) return anim_kid_to_swt(gsm);
      return kid_to_swt(gsm);
    case GSM_TYPE_TAC:
    case GSM_TYPE_MHx:
      SAY("converting TAC/MHx %s to SWT\n", (n) ? "animation" : "model");
      if (n) return anim_tac_to_swt(gsm);
      return tac_to_swt(gsm);
    case GSM_TYPE_BF1:
    case GSM_TYPE_FHx:
      SAY("converting BF1/FHx %s to SWT\n", (n) ? "animation" : "model");
      if (n) return anim_bf1_to_swt(gsm);
      return bf1_to_swt(gsm);
    default:
      error_dbg("don't know how to convert type %u to SWT\n", type);
      return 2;
  }
}

/* stuff that needs to be done for all new SWT models */
int prepare_swt(struct gsm *gsm) {
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  DEBUG("cleaning up original skeleton\n");
  cleanup_skeletons(gsm->num_skeletons, gsm->skeletons, gsm->num_frames);
  gsm->num_skeletons = 0;

  DEBUG("substituting SWT skeleton\n");
  gsm->skeletons = make_swt_skeletons();
  if (! gsm->skeletons) {
    ERROR("can't make SWT skeleton\n");
    return 4;
  }
  gsm->num_skeletons = 1;

  if (gsm->type & GSM_TYPE_ANIMATION) return 0;

  DEBUG("cleaning up original grippers\n");
  cleanup_grippers(gsm->num_grippers, gsm->grippers);
  gsm->num_grippers = 0;

  DEBUG("substituting SWT grippers\n");
  gsm->grippers = make_swt_grippers();
  if (! gsm->grippers) {
    ERROR("can't make SWT grippers\n");
    return 5;
  }
  gsm->num_grippers = SWT_NUM_GRIPPERS;

  return 0;
}

/* convert a 40-bone KID model to SWT */
int kid_to_swt(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_swt */
  if (gsm->type != GSM_TYPE_KID) {
    error_dbg("this is not a KID model, can't convert to SWT\n");
    return 3;
  }

  /* do this first so bone names are right */
  DEBUG("reassigning vertices to new bones\n");
  if (reassign_kid_to_swt(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_swt(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 37-bone TAC/MHx model to SWT */
int tac_to_swt(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_swt */
  if (gsm->type != GSM_TYPE_TAC && gsm->type != GSM_TYPE_MHx) {
    error_dbg("this is not a TAC/BG1/BG2/BG3/MH1/MH2 model, can't convert to SWT\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_tac_to_swt(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_swt(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 34-bone BF1/FHx model to SWT */
int bf1_to_swt(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_swt */
  if (gsm->type != GSM_TYPE_BF1 && gsm->type != GSM_TYPE_FHx) {
    error_dbg("this is not a BF1/FH1/FH2 model, can't convert to SWT\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_bf1_to_swt(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_swt(gsm);
  if (ret) return ret;

  return 0;
}

/* make a 40-bone KID model */
int make_kid(struct gsm *gsm) {
  int n, type;

  n = gsm->type & GSM_TYPE_ANIMATION;
  type = gsm->type % GSM_TYPE_ANIMATION_LIMITED_BONES;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  switch (type) {
    case GSM_TYPE_SWT:
      SAY("converting SWT %s to KID\n", (n) ? "animation" : "model");
      if (n) return anim_swt_to_kid(gsm);
      else return swt_to_kid(gsm);
    case GSM_TYPE_KID:
      error_dbg("GSM is already a KID model\n");
      return 0;
    case GSM_TYPE_TAC:
    case GSM_TYPE_MHx:
      SAY("converting TAC %s to KID\n", (n) ? "animation" : "model");
      if (n) return anim_tac_to_kid(gsm);
      else return tac_to_kid(gsm);
    case GSM_TYPE_BF1:
    case GSM_TYPE_FHx:
      SAY("converting BF1 %s to KID\n", (n) ? "animation" : "model");
      if (n) return anim_bf1_to_kid(gsm);
      else return bf1_to_kid(gsm);
    default:
      error_dbg("don't know how to convert type %u to KID\n", type);
      return 2;
  }
}

/* stuff that needs to be done for all new KID models */
int prepare_kid(struct gsm *gsm) {
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  DEBUG("cleaning up original skeleton\n");
  cleanup_skeletons(gsm->num_skeletons, gsm->skeletons, gsm->num_frames);
  gsm->num_skeletons = 0;

  DEBUG("substituting KID skeleton\n");
  gsm->skeletons = make_kid_skeletons();
  if (! gsm->skeletons) {
    ERROR("can't make KID skeleton\n");
    return 4;
  }
  gsm->num_skeletons = 1;

  if (gsm->type & GSM_TYPE_ANIMATION) return 0;

  DEBUG("cleaning up original grippers\n");
  cleanup_grippers(gsm->num_grippers, gsm->grippers);
  gsm->num_grippers = 0;

  /* KIDs have no grippers */
  return 0;
}

/* convert a 52-bone SWT model to KID */
int swt_to_kid(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_kid */
  if (gsm->type != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT model, can't convert to KID\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_swt_to_kid(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_kid(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 37-bone TAC model to KID */
int tac_to_kid(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_kid */
  if (gsm->type != GSM_TYPE_TAC && gsm->type != GSM_TYPE_MHx) {
    error_dbg("this is not a TAC/BG1/BG2/BG3/MH1/MH2 model, can't convert to KID\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_tac_to_kid(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_kid(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 34-bone BF1 model to KID */
int bf1_to_kid(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_kid */
  if (gsm->type != GSM_TYPE_BF1 && gsm->type != GSM_TYPE_FHx) {
    error_dbg("this is not a BF1/FH1/FH2 model, can't convert to KID\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_bf1_to_kid(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_kid(gsm);
  if (ret) return ret;

  return 0;
}

/* make a 37-bone TAC model */
int make_tac(struct gsm *gsm) {
  int n, type;

  n = gsm->type & GSM_TYPE_ANIMATION;
  type = gsm->type % GSM_TYPE_ANIMATION_LIMITED_BONES;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  switch (type) {
    case GSM_TYPE_SWT:
      SAY("converting SWT %s to TAC\n", (n) ? "animation" : "model");
      if (n) return anim_swt_to_tac(gsm);
      else return swt_to_tac(gsm);
    case GSM_TYPE_KID:
      SAY("converting KID %s to TAC\n", (n) ? "animation" : "model");
      if (n) return anim_kid_to_tac(gsm);
      else return kid_to_tac(gsm);
    case GSM_TYPE_TAC:
      error_dbg("GSM is already a TAC model\n");
      return 0;
    case GSM_TYPE_MHx:
      SAY("converting MHx model to TAC\n");
      return mhx_to_tac(gsm);
    case GSM_TYPE_BF1:
      error_dbg("GSM is already a BF1 model\n");
      return 0;
    case GSM_TYPE_FHx:
      if (n) {
        SAY("compatible animations, nothing to be done\n");
        return 0;
      }
      else {
        SAY("converting FHx model to BF1\n");
        return fhx_to_bf1(gsm);
      }
    default:
      error_dbg("don't know how to convert type %u to TAC\n", type);
      return 2;
  }
}

/* stuff that needs to be done for all new TAC models */
int prepare_tac(struct gsm *gsm) {
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  DEBUG("cleaning up original skeleton\n");
  cleanup_skeletons(gsm->num_skeletons, gsm->skeletons, gsm->num_frames);
  gsm->num_skeletons = 0;

  DEBUG("substituting TAC skeleton\n");
  gsm->skeletons = make_tac_skeletons();
  if (! gsm->skeletons) {
    ERROR("can't make TAC skeleton\n");
    return 4;
  }
  gsm->num_skeletons = 1;

  if (gsm->type & GSM_TYPE_ANIMATION) return 0;

  DEBUG("cleaning up original grippers\n");
  cleanup_grippers(gsm->num_grippers, gsm->grippers);
  gsm->num_grippers = 0;

  DEBUG("substituting TAC grippers\n");
  gsm->grippers = make_tac_grippers();
  if (! gsm->grippers) {
    ERROR("can't make TAC grippers\n");
    return 5;
  }
  gsm->num_grippers = TAC_NUM_GRIPPERS;

  return 0;
}

/* convert a 52-bone SWT model to TAC */
int swt_to_tac(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT model, can't convert to TAC\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_swt_to_tac(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_tac(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 40-bone KID model to TAC */
int kid_to_tac(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_KID) {
    error_dbg("this is not a KID model, can't convert to TAC\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_kid_to_tac(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_tac(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 37-bone TAC/MHx model to SWT */
int mhx_to_tac(struct gsm *gsm) {
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_MHx) {
    error_dbg("this is not a MH1/MH2 model, can't convert to TAC\n");
    return 3;
  }

  DEBUG("adding TAC grippers\n");
  gsm->grippers = make_tac_grippers();
  if (! gsm->grippers) {
    ERROR("can't add TAC grippers\n");
    return 5;
  }
  gsm->num_grippers = TAC_NUM_GRIPPERS;

  return 0;
}

/* make a 34 bone BF1 model */
int make_bf1(struct gsm *gsm) {
  int n, type;

  n = gsm->type & GSM_TYPE_ANIMATION;
  type = gsm->type % GSM_TYPE_ANIMATION_LIMITED_BONES;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  switch (type) {
    case GSM_TYPE_SWT:
      SAY("converting SWT %s to BF1\n", (n) ? "animation" : "model");
      if (n) return anim_swt_to_bf1(gsm);
      else return swt_to_bf1(gsm);
    case GSM_TYPE_KID:
      SAY("converting KID %s to BF1\n", (n) ? "animation" : "model");
      if (n) return anim_kid_to_bf1(gsm);
      else return kid_to_bf1(gsm);
    case GSM_TYPE_TAC:
    case GSM_TYPE_MHx:
      SAY("converting TAC %s to BF1\n", (n) ? "animation" : "model");
      if (n) return anim_tac_to_bf1(gsm);
      else return tac_to_bf1(gsm);
    case GSM_TYPE_BF1:
      if (n) error_dbg("GSA is already a BF1 animation\n");
      error_dbg("GSM is already a BF1 model\n");
      return 0;
    case GSM_TYPE_FHx:
      return make_tac(gsm); /* cheat */
    default:
      error_dbg("don't know how to convert type %u to BF1\n", type);
      return 2;
  }
}

/* stuff that needs to be done for all new BF1 models */
int prepare_bf1(struct gsm *gsm) {
  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  DEBUG("cleaning up original skeleton\n");
  cleanup_skeletons(gsm->num_skeletons, gsm->skeletons, gsm->num_frames);
  gsm->num_skeletons = 0;

  DEBUG("substituting BF1 skeleton\n");
  gsm->skeletons = make_bf1_skeletons();
  if (! gsm->skeletons) {
    ERROR("can't make BF1 skeleton\n");
    return 4;
  }
  gsm->num_skeletons = 1;

  if (gsm->type & GSM_TYPE_ANIMATION) return 0;

  DEBUG("cleaning up original grippers\n");
  cleanup_grippers(gsm->num_grippers, gsm->grippers);
  gsm->num_grippers = 0;

  DEBUG("substituting BF1 grippers\n");
  gsm->grippers = make_bf1_grippers();
  if (! gsm->grippers) {
    ERROR("can't make BF1 grippers\n");
    return 5;
  }
  gsm->num_grippers = BF1_NUM_GRIPPERS;

  return 0;
}

/* convert a 52-bone SWT model to BF1 */
int swt_to_bf1(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT model, can't convert to BF1\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_swt_to_bf1(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_bf1(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 40-bone KID model to BF1 */
int kid_to_bf1(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_KID) {
    error_dbg("this is not a KID model, can't convert to BF1\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_kid_to_bf1(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_bf1(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 37-bone TAC/BGx/MHx model to BF1 */
int tac_to_bf1(struct gsm *gsm) {
  int ret;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_TAC) {
    error_dbg("this is not a TAC model, can't convert to BF1\n");
    return 3;
  }

  DEBUG("reassigning vertices to new bones\n");
  if (reassign_tac_to_bf1(gsm)) {
    ERROR("can't reassign vertices\n");
    return 6;
  }

  ret = prepare_bf1(gsm);
  if (ret) return ret;

  return 0;
}

/* convert a 34-bone FHx model to BF1 */
int fhx_to_bf1(struct gsm *gsm) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  /* return code 2 used by make_tac */
  if (gsm->type != GSM_TYPE_FHx) {
    error_dbg("this is not a FH1/FH2 model, can't convert to BF1\n");
    return 3;
  }

  DEBUG("adding BF1 grippers\n");
  gsm->grippers = make_bf1_grippers();
  if (! gsm->grippers) {
    ERROR("can't add BF1 grippers\n");
    return 5;
  }
  gsm->num_grippers = BF1_NUM_GRIPPERS;

  return 0;
}

/* reset model's bind pose to da Vinci pose */
int make_default_skeleton(struct gsm *gsm) {
  /* pointer to skeleton reassignment function */
  int (*function)(struct gsm *);
  unsigned short type;
  int ret;

  /* check it's a character */
  DEBUG("determining model type");
  switch (gsm->type) {
    case GSM_TYPE_SWT:
      DEBUG("this is a SWT model");
      function = prepare_swt;
      break;
    case GSM_TYPE_TAC:
    case GSM_TYPE_BF1:
    case GSM_TYPE_MHx:
      DEBUG("this is a TAC model");
      function = prepare_tac;
      break;
    case GSM_TYPE_FHx:
      DEBUG("this is a BF1 model");
      function = prepare_bf1;
      break;
    case GSM_TYPE_KID:
      DEBUG("this is a KID model");
      function = prepare_kid;
      break;
    default:
      ERROR("this is not a character model!\n");
      return 1;
  }

  /* HACK: set animation flag so prepare_xxx() doesn't process grippers */
  type = gsm->type;
  gsm->type |= GSM_TYPE_ANIMATION;

  /* call the function */
  DEBUG("calling skeleton preparation function");
  ret = function(gsm);

  /* restore animation flag */
  gsm->type = type;

  if (ret) {
    ERROR("skeleton preparation function returned %u\n", ret);
    return 2;
  }

  return 0;
}

/* convert a character model to a gun(!) */
int character_to_gun(struct gsm *gsm) {
  unsigned short i;
  unsigned short num_grippers;
  struct gripper **grippers;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  /* reassign ALL vertices to the root node */
  DEBUG("clearing all assignments\n");
  cleanup_assignments(gsm->num_assignments, gsm->assignments);
  gsm->num_assignments = 0;
  gsm->assignments = alloc_assignments(1);
  if (! gsm->assignments) return 2;
  gsm->num_assignments = 1;

  DEBUG("making trivial assignments to root node\n");
  gsm->assignments[0]->bone_id = SWT_BIPED_SPINE;
  gsm->assignments[0]->first_vertex = 0;
  gsm->assignments[0]->num_vertices = gsm->num_positions;

  /* pretend grippers are actually grips */
  DEBUG("swapping grips for grippers\n");
  if (gsm->num_grips) {
    SAY("%s already has grips(?)\n", gsm->basename);
  }
  else {
    if (! gsm->num_grippers) {
      SAY("%s has neither grips nor grippers: adding defaults\n");
      ERROR("FIXME\n");
    }
    num_grippers = gsm->num_grippers;
    gsm->num_grippers = gsm->num_grips;
    gsm->num_grips = num_grippers;
    grippers = gsm->grippers;
    gsm->grippers = gsm->grips;
    gsm->grips = grippers;
  }

  /* make a muzzle */
  /* XXX: find material with first position in */
  DEBUG("setting material to %s\n", MUZZLE);
  if (! gsm->num_materials || ! gsm->materials[0]) {
    ERROR("GSM has no materials (why are we even here?)\n");
    return 3;
  }
  if (gsm->materials[0]->name) free (gsm->materials[0]->name);
  gsm->materials[0]->name_length = 0;
  gsm->materials[0]->name = alloc_name(MUZZLE, &gsm->materials[0]->name_length);
  if (! gsm->materials[0]->name_length) {
    ERROR("out of memory for new material name %s", MUZZLE);
    return 4;
  }

  return 0;
}

/*
  add a flashlight to a weapon model
  returns 0 on success, < 0 on error, 1 if weapon already has flashlight
*/
int add_flashlight(struct gsm *gsm) {
  unsigned short i, triangle, triangle_group, num_triangles, num_triangle_groups;
  int material_id;
  struct gll *tmp, *triangle_groups, *triangles, *vertices;
  struct material **new_materials;
  struct triangle **new_triangles;
  struct triangle_group **new_triangle_groups;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return -1;
  }

  if (gsm->type != GSM_TYPE_GUN && gsm->type != GSM_TYPE_PIS) {
    error_dbg("this is not a weapon model, can't add flashlight\n");
    return -2;
  }

  material_id = get_material_by_name(gsm, FLASHLIGHT);
  if (material_id >= 0) {
    ERROR("weapon already has a flashlight\n");
    return 1;
  }
  material_id = get_material_by_name(gsm, MUZZLE);
  if (material_id < 0) {
    ERROR("huh? gsm->type = %u but no " MUZZLE " material??\n", gsm->type);
    return -1;
  }
  else {
    DEBUG(MUZZLE " is material %02u\n", material_id);
  }

  SAY("adding flashlight to %s\n", gsm->filename);

  /* first make the new materials */
  new_materials = alloc_materials(gsm->num_materials + 1);
  DEBUG("making new material\n");
  if (! new_materials) {
    ERROR("out of memory for new materials");
    return -3;
  }
  TRACE("new materials are at %p\n", new_materials);
  for (i = 0; i < gsm->num_materials; i++) {
    TRACE1("memmove({new_materials[%u]=%p}, {gsm->materials[%u]=%p}, {sizeof(struct material)=%u})\n", i, new_materials[i], i, gsm->materials[i], sizeof(struct material));
    memmove(new_materials[i], gsm->materials[i], sizeof(struct material));
    /*
      free each one instead of calling cleanup_materials because we don't 
      want to free the maps as well
    */
    if (i != material_id) free(gsm->materials[i]);
  }

  /* copy all MUZZLE stuff for last material */
  TRACE1("memmove({new_materials[%u]=%p}, {gsm->materials[%u]=%p}, {sizeof(struct material)=%u})\n", i, new_materials[i], material_id, gsm->materials[material_id], sizeof(struct material));
  memmove(new_materials[i], gsm->materials[material_id], sizeof(struct material));
  free(gsm->materials[material_id]);
  /* rename */
  DEBUG("renaming " MUZZLE " to " FLASHLIGHT "\n");
  new_materials[i]->name_length = 0;
//free(new_materials[i]->name);
  new_materials[i]->name = (char *) alloc_name(FLASHLIGHT, &new_materials[i]->name_length);
  new_materials[i]->number = i;
  DEBUG("fixing " FLASHLIGHT " material number to %02u\n", new_materials[i]->number);
  gsm->num_materials++;
  gsm->materials = new_materials;
  DEBUG("old name: %p, new name: %p\n", gsm->materials[material_id]->name, new_materials[i]->name);
  if (! new_materials[i]->name_length) {
    ERROR("out of memory for new name length");
    return -4;
  }
  DEBUG("new material %02u at %p: %s\n", i, gsm->materials[i], gsm->materials[i]->name);

  num_triangles = 0;
  triangles = 0;
  num_triangle_groups = 0;
  triangle_groups = 0;
  vertices = 0;

  /* find triangle groups whose material is MUZZLE */
  DEBUG("finding triangle groups\n");
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    if (gsm->triangle_groups[i]->material_id != material_id) continue;
    DEBUG("triangle group %02u maps " MUZZLE "\n", i);
    triangle_groups = gll_add(&triangle_groups);
    if (! triangle_groups) {
      ERROR("out of memory for triangle group list");
      gll_free(triangle_groups);
      return -4;
    }
    gll_set_last(triangle_groups, i);
    num_triangle_groups++;
  }
  DEBUG("triangle groups mapped to material %02u: %u\n", material_id, num_triangle_groups);
  if (! triangle_groups) {
    SAY(MUZZLE " material exists but is not mapped to any triangles\n");
  }
  gll_debug(triangle_groups);

  /* find triangles in triangle groups */
  DEBUG("finding triangles\n");
  GLL_LOOP(tmp, triangle_groups) {
    triangle_group = tmp->n;
    for (i = 0; i < gsm->triangle_groups[triangle_group]->num_triangles; i++) {
      triangle = gsm->triangle_groups[triangle_group]->first_triangle + i;
      DEBUG("triangle %02u is in triangle group %02u\n", triangle, triangle_group);
      triangles = gll_add(&triangles);
      if (! triangles) {
        ERROR("out of memory for triangle list");
        gll_free(triangles);
        gll_free(triangle_groups);
        return -5;
      }
      gll_set_last(triangles, triangle);
      num_triangles++;
    }
  }
  DEBUG("triangles in group %02u: %u\n", triangle_group, num_triangles);
  if (! triangles) {
    SAY(MUZZLE " triangle groups exist but don't map any triangles!?!\n");
  }
  gll_debug(triangles);

  /* new triangles */
  DEBUG("making new triangles\n");
  new_triangles = alloc_triangles(gsm->num_triangles + num_triangles);
  if (! new_triangles) {
    ERROR("out of memory for new triangles");
    return -6;
  }
  /*
    copy all existing triangles up until the last one that maps to MUZZLE
    we're assuming the GSM is optimised with all triangles correctly grouped
  */
  for (i = 0; i < triangles->n + num_triangles; i++) {
    TRACE1("memmove({new_triangles[%u]=%p}, {gsm->triangles[%u]=%p}, {sizeof(struct triangle)=%u})\n", i, new_triangles[i], i, gsm->triangles[i], sizeof(struct triangle));
    memmove(new_triangles[i], gsm->triangles[i], sizeof(struct triangle));
    DEBUG("copied triangle %02u to %02u\n", i, i);
  }
  /* insert new triangles */
  for (i = 0; i < num_triangles; i++) {
    TRACE1("memmove({new_triangles[%u]=%p}, {gsm->triangles[%u]=%p}, {sizeof(struct triangle)=%u})\n", i + triangles->n + num_triangles, new_triangles[i + triangles->n + num_triangles], i + triangles->n, gsm->triangles[i + triangles->n], sizeof(struct triangle));
    memmove(new_triangles[i + triangles->n + num_triangles], gsm->triangles[i + triangles->n], sizeof(struct triangle));
    DEBUG("copied triangle %02u to %02u\n", i + triangles->n, i + triangles->n + num_triangles);
  }
  /* copy following triangles */
  for (i = triangles->n + num_triangles; i < gsm->num_triangles; i++) {
    TRACE1("memmove({new_triangles[%u]=%p}, {gsm->triangles[%u]=%p}, {sizeof(struct triangle)=%u})\n", i + num_triangles, new_triangles[i + num_triangles], i, gsm->triangles[i], sizeof(struct triangle));
    memmove(new_triangles[i + num_triangles], gsm->triangles[i], sizeof(struct triangle));
    DEBUG("copied triangle %02u to %02u\n", i, i + num_triangles);
  }
  cleanup_triangles(gsm->num_triangles, gsm->triangles);
  gsm->num_triangles += num_triangles;
  gsm->triangles = new_triangles;

  /* then make new triangle group */
  DEBUG("making new triangle groups\n");
  new_triangle_groups = alloc_triangle_groups(gsm->num_triangle_groups + num_triangle_groups);
  if (! new_triangle_groups) {
    ERROR("out of memory for new triangle_groups");
    return -6;
  }
  /* copy preceding existing groups */
  for (i = 0; i < triangle_groups->n + num_triangle_groups; i++) {
    TRACE1("memmove({new_triangle_groups[%u]=%p}, {gsm->triangle_groups[%u]=%p}, {sizeof(struct triangle_group)=%u})\n", i, new_triangle_groups[i], i, gsm->triangle_groups[i], sizeof(struct triangle_group));
    memmove(new_triangle_groups[i], gsm->triangle_groups[i], sizeof(struct triangle_group));
    DEBUG("copied triangle_group %02u to %02u\n", i, i);
  }
  /* insert new groups */
  for (i = 0; i < num_triangle_groups; i++) {
    TRACE1("memmove({new_triangle_groups[%u]=%p}, {gsm->triangle_groups[%u]=%p}, {sizeof(struct triangle_group)=%u})\n", i + triangle_groups->n + num_triangle_groups, new_triangle_groups[i + triangle_groups->n + num_triangle_groups], i + triangle_groups->n, gsm->triangle_groups[i + triangle_groups->n], sizeof(struct triangle_group));
    memmove(new_triangle_groups[i + triangle_groups->n + num_triangle_groups], gsm->triangle_groups[i + triangle_groups->n], sizeof(struct triangle_group));
    /* fix material id */
    new_triangle_groups[i + triangle_groups->n + num_triangle_groups]->material_id = gsm->num_materials - 1;
    /* fix first triangle */
    new_triangle_groups[i + triangle_groups->n + num_triangle_groups]->first_triangle += num_triangles;
    DEBUG("copied triangle_group %02u to %02u\n", i + triangle_groups->n, i + triangle_groups->n + num_triangle_groups);
  }
  /* copy following groups */
  for (i = triangle_groups->n + num_triangle_groups; i < gsm->num_triangle_groups; i++) {
    TRACE1("memmove({new_triangle_groups[%u]=%p}, {gsm->triangle_groups[%u]=%p}, {sizeof(struct triangle_group)=%u})\n", i + num_triangle_groups, new_triangle_groups[i + num_triangle_groups], i, gsm->triangle_groups[i], sizeof(struct triangle_group));
    memmove(new_triangle_groups[i + num_triangle_groups], gsm->triangle_groups[i], sizeof(struct triangle_group));
    /* fix first triangle */
    new_triangle_groups[i + num_triangle_groups]->first_triangle += num_triangles;
    DEBUG("copied triangle_group %02u to %02u\n", i, i + num_triangle_groups);
  }
  cleanup_triangle_groups(gsm->num_triangle_groups, gsm->triangle_groups);
  gsm->num_triangle_groups += num_triangle_groups;
  gsm->triangle_groups = new_triangle_groups;
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    DEBUG("triangle group %u: triangles: %u from %u with material %u\n", i, gsm->triangle_groups[i]->num_triangles, gsm->triangle_groups[i]->first_triangle, gsm->triangle_groups[i]->material_id);
  }

  return 0;
}

/* convert a pistol (weapon, one grip) to a rifle (weapon, two grips) */
int make_rifle(struct gsm *gsm) {
  struct gripper **new_grips;
  unsigned short i;

  if (! gsm) {
    ERROR("called with no GSM!\n");
    return -1;
  }

  if (gsm->type == GSM_TYPE_GUN) {
    error_dbg("GSM is already a rifle\n");
    return 0;
  }
  if (gsm->type != GSM_TYPE_PIS) {
    /*error_dbg("this is not a pistol model, can't convert to rifle\n");*/
    /*return -2;*/
    DEBUG("this is not a pistol model, calling character hack\n");
    return character_to_gun(gsm);
  }
  SAY("converting pistol to rifle\n");

  DEBUG("making new grips\n");
  new_grips = alloc_grippers(GUN_NUM_GRIPS);
  if (! new_grips) {
    ERROR("out of memory for new grips");
    return -3;
  }

  TRACE("new grips are at %p\n", new_grips);
  /* pedantic, could use constant i=0 */
  for (i = 0; i < gsm->num_grips; i++) {
    TRACE1("memmove({new_grips[%u]=%p}, {gsm->grips[%u]=%p}, {sizeof(struct gripper)=%u})\n", i, new_grips[i], i, gsm->grips[i], sizeof(struct gripper));
    memmove(new_grips[i], gsm->grips[i], sizeof(struct gripper));
  }

  /* copy Grip1 */
  TRACE1("memmove({new_grips[%u]=%p}, {gsm->grips[%u]=%p}, {sizeof(struct gripper)=%u})\n", i, new_grips[i], gsm->num_grips - 1, gsm->grips[gsm->num_grips - 1], sizeof(struct gripper));
  memmove(new_grips[i], gsm->grips[gsm->num_grips - 1], sizeof(struct gripper));

  /* rename */
  DEBUG("reallocating first grip name %s\n", gun_grip_name[0]);
  new_grips[0]->name_length = 0;
  new_grips[0]->name = (char *) alloc_name(gun_grip_name[0], &new_grips[0]->name_length);
  if (! new_grips[0]->name_length) {
    ERROR("out of memory for old name length");
    return -4;
  }
  DEBUG("renaming %s to %s\n", new_grips[i]->name, gun_grip_name[gsm->num_grips]);
  new_grips[i]->name_length = 0;
  new_grips[i]->name = (char *) alloc_name(gun_grip_name[gsm->num_grips], &new_grips[i]->name_length);
  if (! new_grips[i]->name_length) {
    ERROR("out of memory for new name length");
    return -5;
  }
  cleanup_grippers(gsm->num_grips, gsm->grips);

  gsm->num_grips++;
  gsm->grips = new_grips;
  DEBUG("new grip %02u at %p: %s\n", i, gsm->grips[i], gsm->grips[i]->name);

  return 0;
}

/* convert arbitrary animations */
int anim_xxx_to_yyy(struct gsm *gsa, struct skeleton **(*new_skeletons_function)(), int (*reassign_function)(unsigned short)) {
  struct skeleton **new_skeletons, *skeleton, *new_skeleton;
  int ret;
  unsigned short i, j;

  if (! gsa) {
    ERROR("called with no GSA for crying out loud!\n");
    return 1;
  }

  if (! reassign_function) {
    ERROR("called with no function!\n");
    return 2;
  }

  DEBUG("making new skeleton\n");
  new_skeletons = new_skeletons_function();
  if (! new_skeletons) {
    ERROR("can't make skeleton\n");
    return 4;
  }

  /* reassign bone IDs in exported bone list */
  skeleton = gsa->skeletons[0];
  new_skeleton = new_skeletons[0];
  if (! skeleton->num_animated_bones) {
    ERROR("huh? no animated bones?\n");
    return 5;
  }
  new_skeleton->num_animated_bones = 0;
  for (i = 0; i < skeleton->num_bones; i++) {
    if (skeleton->num_bones != skeleton->num_animated_bones && ! skeleton->bones[i]->animated) continue;
    ret = reassign_function(i);
    if (ret < 0) {
      ERROR("invalid bone reassignment for bone %u\n", i);
      return 6;
    }
    if (ret == BONE_FIRST_LINK) {
      DEBUG("animated bone %u has no match in new skeleton\n", i);
      continue;
    }
    DEBUG("marking bone %d in new skeleton for animation\n", ret);
    new_skeleton->bones[ret]->animated = 1;
    DEBUG("making new orientations for bone %d\n", ret);
    new_skeleton->bones[ret]->orientations = alloc_quaternions(gsa->num_frames);
    if (! new_skeleton->bones[ret]->orientations) {
      ERROR("out of memory for new orientations");
      return 7;
    }
    /* allocate matrices too */
    new_skeleton->bones[ret]->locals = alloc_matrices(gsa->num_frames);
    if (! new_skeleton->bones[ret]->locals) {
      ERROR("out of memory for new local rotation matrices");
      return 8;
    }
    new_skeleton->bones[ret]->worlds = alloc_matrices(gsa->num_frames);
    if (! new_skeleton->bones[ret]->worlds) {
      ERROR("out of memory for new world rotation matrices");
      return 9;
    }
    for (j = 0; j < gsa->num_frames; j++) {
      TRACE2("copying orientation %u\n", j);
      copy_quaternion(new_skeleton->bones[ret]->orientations[j], skeleton->bones[i]->orientations[j]);
      copy_matrix(new_skeleton->bones[ret]->locals[j], skeleton->bones[i]->locals[j]);
      copy_matrix(new_skeleton->bones[ret]->worlds[j], skeleton->bones[i]->worlds[j]);
    }
    new_skeleton->num_animated_bones++;
  }

  DEBUG("replacing skeleton\n");
  free_structs("skeleton", gsa->num_skeletons, (void *) gsa->skeletons);
  gsa->skeletons = new_skeletons;
  gsa->num_skeletons = 1;

  return 0;
}

int anim_kid_to_swt(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_KID) {
    error_dbg("this is not a KID animation, can't convert to SWT\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_swt_skeletons, reassign_bone_id_kid_to_swt);
}

int anim_tac_to_swt(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_TAC && gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_MHx) {
    error_dbg("this is not a TAC/BG1/BG2/BG3/MH1/MH2 animation, can't convert to SWT\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_swt_skeletons, reassign_bone_id_tac_to_swt);
}

int anim_bf1_to_swt(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_BF1) {
    error_dbg("this is not a BF1/FH1/FH2 animation, can't convert to SWT\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_swt_skeletons, reassign_bone_id_bf1_to_swt);
}

int anim_swt_to_kid(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT animation, can't convert to KID\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_kid_skeletons, reassign_bone_id_swt_to_kid);
}

int anim_tac_to_kid(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_TAC) {
    error_dbg("this is not a TAC animation, can't convert to KID\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_kid_skeletons, reassign_bone_id_tac_to_kid);
}

int anim_bf1_to_kid(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_BF1) {
    error_dbg("this is not a BF1 animation, can't convert to KID\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_kid_skeletons, reassign_bone_id_bf1_to_kid);
}

int anim_swt_to_tac(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT animation, can't convert to TAC\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_tac_skeletons, reassign_bone_id_swt_to_tac);
}

int anim_kid_to_tac(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_KID) {
    error_dbg("this is not a KID animation, can't convert to TAC\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_tac_skeletons, reassign_bone_id_kid_to_tac);
}

int anim_swt_to_bf1(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_SWT) {
    error_dbg("this is not a SWT animation, can't convert to BF1\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_bf1_skeletons, reassign_bone_id_swt_to_bf1);
}

int anim_kid_to_bf1(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_KID) {
    error_dbg("this is not a KID animation, can't convert to TAC\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_bf1_skeletons, reassign_bone_id_kid_to_bf1);
}

int anim_tac_to_bf1(struct gsm *gsa) {
  if (! gsa) {
    ERROR("called with no GSA!\n");
    return 1;
  }

  if (gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_TAC && gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES != GSM_TYPE_MHx) {
    error_dbg("this is not a TAC/BG1/BG2/BG3/MH1/MH2 animation, can't convert to BF1\n");
    return 3;
  }

  return anim_xxx_to_yyy(gsa, make_bf1_skeletons, reassign_bone_id_tac_to_bf1);
}
