#ifndef MATHS_H
#define MATHS_H
#include <math.h>

/* don't need it but it may be helpful to have around */
#ifdef M_PI
#define PI M_PI
#else
#define PI ((double) (355 / 133))
#endif

#define FLOAT_EQUIV_THRESHOLD 0.0001

typedef struct vector {
  float x;
  float y;
  float z;
} vector_t;

typedef struct quaternion {
  float x;
  float y;
  float z;
  float w;
} quaternion_t;

typedef float matrix_t[16];
typedef float euler_t[3];

/*
  matrices are 4x4 
  indices as per OpenGL:

   | 0  4  8 12 |       |  r  r  r tx |   r = rotation       |  0  3  6  . |
   | 1  5  9 13 |  -->  |  r  r  r ty |  s* = scale          |  1  4  7  . |
   | 2  6 10 14 |  -->  |  r  r  r tz |  t* = translation    |  2  5  8  . |
   | 3  7 11 15 |       | sx sy sz gs |  gs = global scale   |  .  .  .  . |

*/
/* constants for accessing the matrix */
#define R0 0
#define R1 1
#define R2 2
#define R3 4
#define R4 5
#define R5 6
#define R6 8
#define R7 9
#define R8 10
#define TX 12
#define TY 13
#define TZ 14

/* handle matrix pointers without a sore head */
#define M (*m)
#define N (*n)
#define copy_matrix(x,y) memmove(x,y, sizeof(matrix_t))
#define copy_vector(x,y) memmove(x,y, sizeof(vector_t))
#define copy_quaternion(x,y) memmove(x,y, sizeof(quaternion_t))

GSM_API vector_t **alloc_vectors(unsigned short);
GSM_API vector_t *alloc_vector();
GSM_API void cleanup_vectors(unsigned short, vector_t **);
GSM_API matrix_t **alloc_matrices(unsigned short);
GSM_API matrix_t *alloc_matrix();
GSM_API void cleanup_matrices(unsigned short, matrix_t **);
GSM_API vector_t *scalar_product(float, vector_t *);
GSM_API float dot_product(vector_t *, vector_t *);
GSM_API vector_t *cross_product(vector_t *, vector_t *);
GSM_API float vector_magnitude(vector_t *);
GSM_API int normalise_vector(vector_t *);
GSM_API int vector_equiv(vector_t *, vector_t *);
GSM_API vector_t *get_vector_by_positions(struct position *, struct position *);
GSM_API matrix_t *make_rotation_matrix_in_x(float);
GSM_API matrix_t *make_rotation_matrix_in_y(float);
GSM_API matrix_t *make_rotation_matrix_in_z(float);
GSM_API int quaternion_to_matrix(matrix_t *, quaternion_t *);
GSM_API int transpose_matrix(matrix_t *);
GSM_API void debug_matrix(matrix_t *);
GSM_API int matrix_to_euler(euler_t *, matrix_t *);
GSM_API void debug_euler(euler_t *);
GSM_API void debug_vector(vector_t *);
GSM_API float quaternion_magnitude(quaternion_t *);
GSM_API int normalise_quaternion(quaternion_t *);
GSM_API quaternion_t *multiply_quaternions(quaternion_t *, quaternion_t *);
GSM_API quaternion_t *concatenate_quaternions(quaternion_t *, quaternion_t *);
GSM_API matrix_t *multiply_matrices(matrix_t *, matrix_t *);
GSM_API int invert_quaternion(quaternion_t *);
GSM_API vector_t *multiply_vector_by_matrix(vector_t *, matrix_t *);
GSM_API int quaternion_to_euler(euler_t *, quaternion_t *);
GSM_API int euler_to_quaternion(quaternion_t *, euler_t *);
GSM_API int euler_to_matrix(matrix_t *, euler_t *);
GSM_API int matrix_to_quaternion(quaternion_t *, matrix_t *);
GSM_API void debug_quaternion(quaternion_t *);
GSM_API quaternion_t **alloc_quaternions(unsigned short);
GSM_API quaternion_t *alloc_quaternion();
GSM_API void cleanup_quaternions(unsigned short, quaternion_t **);
GSM_API quaternion_t *multiply_quaternion_by_vector(quaternion_t *, vector_t *);
GSM_API quaternion_t *multiply_vector_by_quaternion(vector_t *, quaternion_t *);
GSM_API vector_t *rotate_vector(vector_t *, quaternion_t *);
GSM_API float radians_to_degrees(float);
GSM_API float degrees_to_radians(float);
GSM_API quaternion_t *find_rotation(quaternion_t *, quaternion_t *);
GSM_API int set_matrix_translation(matrix_t *, vector_t *);
GSM_API int read_euler(FILE *, euler_t *);
GSM_API int read_vector(FILE *, vector_t *);
GSM_API int read_quaternion(FILE *, quaternion_t *);
GSM_API int read_matrix(FILE *, matrix_t *);
GSM_API int write_euler(FILE *, euler_t *);
GSM_API int write_vector(FILE *, vector_t *);
GSM_API int write_quaternion(FILE *, quaternion_t *);
GSM_API int write_matrix(FILE *, matrix_t *);

#endif
