#include "gsmhacking.h"

ms3d_model_t *alloc_ms3d_model() {
  ms3d_model_t *ret = alloc_one("ms3d_model", sizeof(ms3d_model_t));
  if (! ret) return 0;
  memmove(&ret->header.magic, MS3D_MAGIC, sizeof(ret->header.magic));
  ret->header.version = MS3D_VERSION;
  return ret;
}

void cleanup_ms3d_model(ms3d_model_t *model) {
  unsigned short i;

  if (! model) {
    ERROR("no model!\n");
    return;
  }
  free_structs("ms3d_vertex", model->num_vertices, (void *) model->vertices);
  free_structs("ms3d_triangle", model->num_triangles, (void *) model->triangles);
  for (i = 0; i < model->num_groups; i++) cleanup_ms3d_group(model->groups[i]);
  free_structs("ms3d_material", model->num_materials, (void *) model->materials);
  free_structs("ms3d_joint", model->num_joints, (void *) model->joints);
}

ms3d_vertex_t **alloc_ms3d_vertices(unsigned short n) {
  return (ms3d_vertex_t **) alloc_lots(n, "ms3d_vertex", sizeof(ms3d_vertex_t *), (void *(*)()) alloc_ms3d_vertex);
}

ms3d_vertex_t *alloc_ms3d_vertex() {
  return (ms3d_vertex_t *) alloc_one("ms3d_vertex", sizeof(ms3d_vertex_t));
}

ms3d_triangle_t **alloc_ms3d_triangles(unsigned short n) {
  return (ms3d_triangle_t **) alloc_lots(n, "ms3d_triangle", sizeof(ms3d_triangle_t *), (void *(*)()) alloc_ms3d_triangle);
}

ms3d_triangle_t *alloc_ms3d_triangle() {
  return (ms3d_triangle_t *) alloc_one("ms3d_triangle", sizeof(ms3d_triangle_t));
}

ms3d_group_t *alloc_ms3d_group() {
  return (ms3d_group_t *) alloc_one("ms3d_group", sizeof(ms3d_group_t));
}

ms3d_group_t **alloc_ms3d_groups(unsigned short n) {
  return (ms3d_group_t **) alloc_lots(n, "ms3d_group", sizeof(ms3d_group_t *), (void *(*)()) alloc_ms3d_group);
}

void cleanup_ms3d_group(ms3d_group_t *group) {
  if (! group) {
    ERROR("no group!|n");
    return;
  }
  free_structs("triangle_id", group->num_triangles, (void *) group->triangle_id);
  free(group);
}

ms3d_material_t **alloc_ms3d_materials(unsigned short n) {
  return (ms3d_material_t **) alloc_lots(n, "ms3d_material", sizeof(ms3d_material_t *), (void *(*)()) alloc_ms3d_material);
}

ms3d_material_t *alloc_ms3d_material() {
  return (ms3d_material_t *) alloc_one("ms3d_material", sizeof(ms3d_material_t));
}

ms3d_keyframe_rot_t **alloc_ms3d_rotation_keyframes(unsigned short n) {
  return (ms3d_keyframe_rot_t **) alloc_lots(n, "ms3d_keyframe_rot", sizeof(ms3d_keyframe_rot_t *), (void *(*)()) alloc_ms3d_rotation_keyframe);
}

ms3d_keyframe_rot_t *alloc_ms3d_rotation_keyframe() {
  return (ms3d_keyframe_rot_t *) alloc_one("ms3d_keyframe_rot", sizeof(ms3d_keyframe_rot_t));
}

ms3d_keyframe_pos_t **alloc_ms3d_position_keyframes(unsigned short n) {
  return (ms3d_keyframe_pos_t **) alloc_lots(n, "ms3d_keyframe_pos", sizeof(ms3d_keyframe_pos_t *), (void *(*)()) alloc_ms3d_position_keyframe);
}

ms3d_keyframe_pos_t *alloc_ms3d_position_keyframe() {
  return (ms3d_keyframe_pos_t *) alloc_one("ms3d_keyframe_pos", sizeof(ms3d_keyframe_pos_t));
}

ms3d_joint_t **alloc_ms3d_joints(unsigned short n) {
  return (ms3d_joint_t **) alloc_lots(n, "ms3d_joint", sizeof(ms3d_joint_t *), (void *(*)()) alloc_ms3d_joint);
}

ms3d_joint_t *alloc_ms3d_joint() {
  return (ms3d_joint_t *) alloc_one("ms3d_joint", sizeof(ms3d_joint_t));
}

/* .ms3d files have bone/material etc names limited to 32 characters */
int ms3d_truncate_string(char *dest, char *src, int max_length) {
  unsigned short i;
  int length;

  if (! dest) {
    ERROR("no destination buffer\n");
    return -1;
  }
  if (! src) {
    memset(dest, 0, max_length);
    return 0;
  }

  length = strlen(src);
  if (length >= max_length) {
    WARN("name \"%s\" is too long for Milkshape (max %d characters)\n", src, max_length);
    length = max_length - 1;
  }
  for (i = 0; i < max_length; i++) {
    if (i < length) dest[i] = *(src + i);
    else dest[i] = '\0';
  }

  return length + 1;
}

int read_ms3d_file(ms3d_model_t *model) {
  if (! model || ! model->filename) {
    ERROR("no filename to open!\n");
    return 1;
  }
  model->fd = open_read(model->filename);
  if (! model->fd) {
    ERROR("can't open file %s!\n", model->filename);
    return 2;
  }
  DEBUG("opened %s on descriptor %p\n", model->filename, model->fd);

  return 0;
}

int read_ms3d_data(ms3d_model_t *model) {
  unsigned int pos;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  if (read_ms3d_file(model)) {
    ERROR("failed to open file descriptor for %s\n", model->basename);
    return 2;
  }

  INFO("reading Milkshape .ms3d file\n");
  if (read_ms3d_header(model)) return 3;
  if (check_ms3d_magic(model)) return 4;
  if (read_ms3d_vertices(model)) return 5;
  if (read_ms3d_triangles(model)) return 6;
  if (read_ms3d_groups(model)) return 7;
  if (read_ms3d_materials(model)) return 8;
  if (read_ms3d_animation_stuff(model)) return 9;
  if (read_ms3d_joints(model)) return 10;

  /* check we're at end of file */
  pos = get_file_position(model->fd);
  if (fseek(model->fd, 0, SEEK_END)) {
    ERROR("can't seek to end of file to check if we've read it all");
    return 0;
  }
  if (get_file_position(model->fd) != pos) WARN("read last joint but not at end of file: offset 0x%04x, end: 0x%04x\n", pos, get_file_position(model->fd));

  return 0;
}

int read_ms3d_header(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading .ms3d header\n");
  if (read_arbitrary(model->fd, model->header.magic, 10)) {
    ERROR("can't read magic\n");
    return 2;
  }
  if (read_int(model->fd, &model->header.version)) {
    ERROR("can't read version\n");
    return 3;
  }
  DEBUG("magic \"%s\" version %d\n", model->header.magic, model->header.version);

  return 0;
}

int check_ms3d_magic(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("checking .ms3d magic\n");
  if (! str_equals(model->header.magic, MS3D_MAGIC)) {
    error_dbg("incorrect file header: this is not a Milkshape .ms3d file\n");
    return 2;
  }
  if (model->header.version != MS3D_VERSION) {
    WARN("version mismatch: expected version %d, got version %d\n", MS3D_VERSION, model->header.version);
  }

  return 0;
}

int read_ms3d_vertices(ms3d_model_t *model) {
  unsigned short i;
  ms3d_vertex_t *vertex;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading number of vertices\n");
  if (read_short(model->fd, &model->num_vertices)) {
    ERROR("can't read number of vertices\n");
    return 2;
  }
  model->vertices = alloc_ms3d_vertices(model->num_vertices);
  if (! model->vertices) {
    ERROR("out of memory for vertices");
    model->num_vertices = 0;
    return 3;
  }

  for (i = 0; i < model->num_vertices; i++) {
    vertex = model->vertices[i];
    /* flags */
    if (read_unsigned_char(model->fd, &vertex->flags)) {
      ERROR("can't read vertex %u flags\n", i);
      return 4;
    }
    /* position */
    if (read_vector(model->fd, &vertex->position)) {
      ERROR("can't read vertex %u position\n", i);
      return 5;
    }
    /* bone_id */
    if (read_char(model->fd, &vertex->bone_id)) {
      ERROR("can't read vertex %u bone ID\n", i);
      return 6;
    }
    /* reference_count */
    if (read_unsigned_char(model->fd, &vertex->reference_count)) {
      ERROR("can't read vertex %u reference count\n", i);
      return 7;
    }

    TRACE("vertex %u position (% 0.4f, % 0.4f, % 0.4f) linked to bone %d with flags %u, reference count %d\n", i, vertex->position.x, vertex->position.y, vertex->position.z, vertex->bone_id, vertex->flags, vertex->reference_count);
  }
  DEBUG("read %u %s\n", i, (i == 1) ? "vertex" : "vertices");

  return 0;
}

int read_ms3d_triangles(ms3d_model_t *model) {
  unsigned short i;
  ms3d_triangle_t *triangle;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading number of triangles\n");
  if (read_short(model->fd, &model->num_triangles)) {
    ERROR("can't read number of triangles\n");
    return 2;
  }
  model->triangles = alloc_ms3d_triangles(model->num_triangles);
  if (! model->triangles) {
    ERROR("out of memory for triangles");
    model->num_triangles = 0;
    return 3;
  }

  for (i = 0; i < model->num_triangles; i++) {
    triangle = model->triangles[i];
    /* flags */
    if (read_short(model->fd, &triangle->flags)) {
      ERROR("can't read triangle %u flags\n", i);
      return 4;
    }
    /* vertex indices */
    if (read_short(model->fd, &triangle->vertex_id[0])) {
      ERROR("can't read triangle %u first vertex\n", i);
      return 5;
    }
    if (read_short(model->fd, &triangle->vertex_id[1])) {
      ERROR("can't read triangle %u second vertex\n", i);
      return 6;
    }
    if (read_short(model->fd, &triangle->vertex_id[2])) {
      ERROR("can't read triangle %u third vertex\n", i);
      return 7;
    }
    /* normals */
    if (read_vector(model->fd, &triangle->normals[0])) {
      ERROR(" can't read triangle %u first normal\n", i);
      return 8;
    }
    if (read_vector(model->fd, &triangle->normals[1])) {
      ERROR(" can't read triangle %u second normal\n", i);
      return 9;
    }
    if (read_vector(model->fd, &triangle->normals[2])) {
      ERROR(" can't read triangle %u third normal\n", i);
      return 10;
    }
    /* s */
    if (read_float(model->fd, &triangle->s[0])) {
      ERROR("can't read triangle %u first s\n", i);
      return 11;
    }
    if (read_float(model->fd, &triangle->s[1])) {
      ERROR("can't read triangle %u second s\n", i);
      return 12;
    }
    if (read_float(model->fd, &triangle->s[2])) {
      ERROR("can't read triangle %u third s\n", i);
      return 13;
    }
    /* t */
    if (read_float(model->fd, &triangle->t[0])) {
      ERROR("can't read triangle %u first t\n", i);
      return 14;
    }
    if (read_float(model->fd, &triangle->t[1])) {
      ERROR("can't read triangle %u second t\n", i);
      return 15;
    }
    if (read_float(model->fd, &triangle->t[2])) {
      ERROR("can't read triangle %u third t\n", i);
      return 16;
    }
    /* smoothing group */
    if (read_unsigned_char(model->fd, &triangle->smoothing_group)) {
      ERROR("can't read triangle %u smoothing group\n", i);
      return 17;
    }
    /* group_id */
    if (read_unsigned_char(model->fd, &triangle->group_id)) {
      ERROR("can't read triangle %u group ID\n", i);
      return 18;
    }

    TRACE("triangle %u vertices (%u, %u, %u) with flags %u, group %d; UVs: { (%0.4f, %0.4f), (%0.4f, %0.4f), (%0.4f, %0.4f) }\n", i, triangle->vertex_id[0], triangle->vertex_id[1], triangle->vertex_id[2], triangle->flags, triangle->group_id, triangle->s[0], triangle->t[0], triangle->s[1], triangle->t[1], triangle->s[2], triangle->t[2]);
  }
  DEBUG("read %u triangle%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int read_ms3d_groups(ms3d_model_t *model) {
  unsigned short i, j;
  ms3d_group_t *group;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading number of groups\n");
  if (read_short(model->fd, &model->num_groups)) {
    ERROR("can't read number of groups\n");
    return 2;
  }
  model->groups = alloc_ms3d_groups(model->num_groups);
  if (! model->groups) {
    ERROR("out of memory for groups");
    model->num_groups = 0;
    return 3;
  }

  for (i = 0; i < model->num_groups; i++) {
    group = model->groups[i];
    /* flags */
    if (read_unsigned_char(model->fd, &group->flags)) {
      ERROR("can't read group %u flags\n", i);
      return 4;
    }
    /* name */
    if (read_arbitrary(model->fd, &group->name, sizeof(group->name))) {
      ERROR("can't read group %u name\n", i);
      return 5;
    }
    /* number of triangles */
    if (read_short(model->fd, &group->num_triangles)) {
      ERROR("can't read group %u number of triangles\n", i);
      return 6;
    }
    /* allocate enough shorts for the indices */
    group->triangle_id = (unsigned short **) alloc_lots(group->num_triangles, "triangle_id", sizeof(unsigned short *), (void *(*)()) alloc_short);
    if (! group->triangle_id) {
      ERROR("out of memory for group %u triangle IDs\n", i);
      group->num_triangles = 0;
      return 7;
    }
    /* read them in */
    for (j = 0; j < group->num_triangles; j++) {
      if (read_short(model->fd, group->triangle_id[j])) {
        ERROR("can't read group %u triangle %u\n", i, j);
        return 8;
      }
      TRACE1("group %u triangle %u: %u\n", i, j, *group->triangle_id[j]);
    }
    /* material_id */
    if (read_char(model->fd, &group->material_id)) {
      ERROR("can't read group %u material ID\n", i);
      return 9;
    }

    TRACE("group %u \"%s\" has %u triangle%s with material index %d, flags %u\n", i, group->name, group->num_triangles, (group->num_triangles == 1) ? "" : "s", group->material_id, group->flags); 
  }
  DEBUG("read %u group%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int read_ms3d_materials(ms3d_model_t *model) {
  unsigned short i;
  ms3d_material_t *material;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading number of materials\n");
  if (read_short(model->fd, &model->num_materials)) {
    ERROR("can't read number of materials\n");
    return 2;
  }
  model->materials = alloc_ms3d_materials(model->num_materials);
  if (! model->materials) {
    ERROR("out of memory for materials");
    model->num_materials = 0;
    return 3;
  }

  for (i = 0; i < model->num_materials; i++) {
    material = model->materials[i];
    /* name */
    if (read_arbitrary(model->fd, &material->name, sizeof(material->name))) {
      ERROR("can't read material %u name\n", i);
      return 4;
    }
    /* ambient */
    if (read_float(model->fd, &material->ambient[0])) {
      ERROR("can't read material %u ambient red\n", i);
      return 5;
    }
    if (read_float(model->fd, &material->ambient[1])) {
      ERROR("can't read material %u ambient green\n", i);
      return 6;
    }
    if (read_float(model->fd, &material->ambient[2])) {
      ERROR("can't read material %u ambient blue\n", i);
      return 7;
    }
    if (read_float(model->fd, &material->ambient[3])) {
      ERROR("can't read material %u ambient alpha\n", i);
      return 8;
    }
    /* diffuse */
    if (read_float(model->fd, &material->diffuse[0])) {
      ERROR("can't read material %u diffuse red\n", i);
      return 9;
    }
    if (read_float(model->fd, &material->diffuse[1])) {
      ERROR("can't read material %u diffuse green\n", i);
      return 10;
    }
    if (read_float(model->fd, &material->diffuse[2])) {
      ERROR("can't read material %u diffuse blue\n", i);
      return 11;
    }
    if (read_float(model->fd, &material->diffuse[3])) {
      ERROR("can't read material %u diffuse alpha\n", i);
      return 12;
    }
    /* specular */
    if (read_float(model->fd, &material->specular[0])) {
      ERROR("can't read material %u specular red\n", i);
      return 13;
    }
    if (read_float(model->fd, &material->specular[1])) {
      ERROR("can't read material %u specular green\n", i);
      return 14;
    }
    if (read_float(model->fd, &material->specular[2])) {
      ERROR("can't read material %u specular blue\n", i);
      return 15;
    }
    if (read_float(model->fd, &material->specular[3])) {
      ERROR("can't read material %u specular alpha\n", i);
      return 16;
    }
    /* emissive */
    if (read_float(model->fd, &material->emissive[0])) {
      ERROR("can't read material %u emissive red\n", i);
      return 17;
    }
    if (read_float(model->fd, &material->emissive[1])) {
      ERROR("can't read material %u emissive green\n", i);
      return 18;
    }
    if (read_float(model->fd, &material->emissive[2])) {
      ERROR("can't read material %u emissive blue\n", i);
      return 19;
    }
    if (read_float(model->fd, &material->emissive[3])) {
      ERROR("can't read material %u emissive alpha\n", i);
      return 20;
    }
    /* shininess */
    if (read_float(model->fd, &material->shininess)) {
      ERROR("can't read material %u shininess\n", i);
      return 21;
    }
    /* opacity */
    if (read_float(model->fd, &material->opacity)) {
      ERROR("can't read material %u opacity\n", i);
      return 22;
    }
    /* mode */
    if (read_char(model->fd, &material->mode)) {
      ERROR("can't read material %u mode\n", i);
      return 23;
    }
    /* bitmap */
    if (read_arbitrary(model->fd, &material->texture, sizeof(material->texture))) {
      ERROR("can't read material %u texture name\n", i);
      return 24;
    }
    /* alpha map */
    if (read_arbitrary(model->fd, &material->alphamap, sizeof(material->alphamap))) {
      ERROR("can't read material %u alpha map name\n", i);
      return 25;
    }

    TRACE("material %u \"%s\" opacity %d, mode %d\n", i, material->texture, (int) (material->opacity * 100), material->mode);
  }
  DEBUG("read %u material%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int read_ms3d_animation_stuff(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  /* framerate */
  if (read_float(model->fd, &model->framerate)) {
    ERROR("can't read framerate\n");
    return 2;
  }
  DEBUG("framerate: %f fps\n", model->framerate);
  /* current time */
  if (read_float(model->fd, &model->current_time)) {
    ERROR("can't read time\n");
    return 3;
  }
  DEBUG("time: %f\n", model->current_time);
  /* frames */
  if (read_int(model->fd, &model->num_frames)) {
    ERROR("can't read number of frames\n");
    return 4;
  }
  DEBUG("frames: %d\n", model->num_frames);

  return 0;
}

int read_ms3d_joints(ms3d_model_t *model) {
  unsigned short i, j;
  ms3d_joint_t *joint;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("reading number of joints\n");
  if (read_short(model->fd, &model->num_joints)) {
    ERROR("can't read number of joints\n");
    return 2;
  }
  model->joints = alloc_ms3d_joints(model->num_joints);
  if (! model->joints) {
    ERROR("out of memory for joints\n");
    model->num_joints = 0;
    return 3;
  }

  for (i = 0; i < model->num_joints; i++) {
    joint = model->joints[i];
    /* flags */
    if (read_unsigned_char(model->fd, &joint->flags)) {
      ERROR("can't read joint %u flags\n", i);
      return 4;
    }
    /* name */
    if (read_arbitrary(model->fd, &joint->name, sizeof(joint->name))) {
      ERROR("can't read joint %u name\n", i);
      return 5;
    }
    /* parent's name */
    if (read_arbitrary(model->fd, &joint->parent_name, sizeof(joint->parent_name))) {
      ERROR("can't read joint %u parent name\n", i);
      return 6;
    }
    /* rotation */
    if (read_euler(model->fd, &joint->rotation)) {
      ERROR("can't read joint %u rotation\n", i);
      return 7;
    }
    DEBUG("read rotation euler\n");
    debug_euler(&joint->rotation);
    /* position */
    if (read_vector(model->fd, &joint->position)) {
      ERROR("can't read joint %u position\n", i);
      return 8;
    }
    debug_vector(&joint->position);
    /* rotation keyframes */
    if (read_short(model->fd, &joint->num_rotation_keyframes)) {
      ERROR("can't read joint %u number of rotation keyframes\n", i);
      return 9;
    }
    /* position keyframes */
    if (read_short(model->fd, &joint->num_position_keyframes)) {
      ERROR("can't read joint %u number of position keyframes\n", i);
      return 10;
    }
    /* allocate enough rotation structs */
    joint->rotation_keyframes = alloc_ms3d_rotation_keyframes(joint->num_rotation_keyframes);
    if (! joint->rotation_keyframes) {
      ERROR("out of memory for joint %u rotation keyframes", i);
      joint->num_rotation_keyframes = joint->num_position_keyframes = 0;
      return 11;
    }
    /* read them in */
    for (j = 0; j < joint->num_rotation_keyframes; j++) {
      if (read_float(model->fd, &joint->rotation_keyframes[j]->time)) {
        ERROR("can't read joint %u rotation keyframe %u time\n", i, j);
        return 12;
      }
      if (read_euler(model->fd, &joint->rotation_keyframes[j]->rotation)) {
        ERROR("can't read joint %u rotation keyframe %u rotation\n", i, j);
        return 13;
      }
      TRACE1("joint %u rotation keyframe %u: (% 0.4f, % 0.4f, % 0.4f) at %f\n", i, j, joint->rotation_keyframes[j]->rotation[0], joint->rotation_keyframes[j]->rotation[1], joint->rotation_keyframes[j]->rotation[2]);
    }
    /* allocate enough position structs */
    joint->position_keyframes = alloc_ms3d_position_keyframes(joint->num_position_keyframes);
    if (! joint->position_keyframes) {
      ERROR("out of memory for joint %u position keyframes", i);
      joint->num_position_keyframes = 0;
      return 14;
    }
    /* read them in */
    for (j = 0; j < joint->num_position_keyframes; j++) {
      if (read_float(model->fd, &joint->position_keyframes[j]->time)) {
        ERROR("can't read joint %u position keyframe %u time\n", i, j);
        return 15;
      }
      if (read_vector(model->fd, &joint->position_keyframes[j]->position)) {
        ERROR("can't read joint %u position keyframe %u position\n", i, j);
        return 16;
      }
      TRACE1("joint %u position keyframe %u: (% 0.4f, % 0.4f, % 0.4f) at %f\n", i, j, joint->position_keyframes[j]->position.x, joint->position_keyframes[j]->position.y, joint->position_keyframes[j]->position.z);
    }

    TRACE("joint %u \"%s\" linked to \"%s\" position (% 0.4f, % 0.4f, % 0.4f) rotation (% 0.4f, % 0.4f, % 0.4f) with flags %u, rotation keyframes: %u, position keyframes: %u\n", i, joint->name, joint->parent_name, joint->position.x, joint->position.y, joint->position.z, joint->rotation[0], joint->rotation[1], joint->rotation[2], joint->flags, joint->num_rotation_keyframes, joint->num_position_keyframes);
  }
  DEBUG("read %u joint%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int ms3d_to_gsm(struct gsm *gsm, ms3d_model_t *model) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  if (! model->num_joints) {
    ERROR("model has no bones\n");
    return 3;
  }

  SAY("converting Milkshape 3D data to GSkin format\n");
  INFO("setting version identifier\n");
  gsm->version = 1029;
  if (ms3d_to_gsm_process_vertices(gsm, model)) return 3;
  if (ms3d_to_gsm_process_triangles(gsm, model)) return 4;
  if (ms3d_to_gsm_process_groups(gsm, model)) return 5;
  if (ms3d_to_gsm_process_materials(gsm, model)) return 6;
  if (ms3d_to_gsm_process_joints(gsm, model)) return 7;
  if (ms3d_to_gsm_create_grips(gsm, model)) return 8;
  /* FIXME: work out world co-ords */
  if (make_shared_triangle_list(gsm)) return 9;

  return 0;
}

/*
  make positions and vertices
  each position has co-ords of the ms3d_vertex
  make one vertex for each position - we'll fill in texture co-ords later
*/
int ms3d_to_gsm_process_vertices(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i, j, n, assignment_id, count;
  struct position *position;
  struct vertex *vertex;
  struct assignment *assignment;
  int dup, last_bone;
  ms3d_vertex_t *v, *u;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  DEBUG("setting positions from vertices\n");
  /* all vertices must be assigned */
  for (i = 0; i < model->num_vertices; i++) {
    if (model->vertices[i]->bone_id > -1) count++;
  }
  if (! count) {
    SAY("no vertices have been assigned to bones\n");
    SAY("assuming all vertices should be assigned to the root node\n");
    for (i = 0; i < model->num_vertices; i++) {
      model->vertices[i]->bone_id = 0;
    }
  }
  else if (count < model->num_vertices) {
    ERROR("some vertices were not assigned to a bone - can't use this model\n");
    return -1;
  }

  gsm->positions = alloc_positions(model->num_vertices);
  if (! gsm->positions) {
    ERROR("out of memory for positions");
    return 3;
  }
  gsm->num_positions = model->num_vertices;

  n = 0; /* new position index */
  for (i = 0; i < model->num_vertices; i++) {
    v = model->vertices[i];
    if (! v->reference_count) {
      WARN("vertex %u has reference count zero; fix your model!\n", i);
    }
    /* check for dups */
    dup = 0;
    for (j = 0; j < i; j++) {
      u = model->vertices[j];
      if (vector_equiv(&v->position, &u->position)) {
        DEBUG("duplicate vertex %u = %u (% 0.4f, % 0.4f, % 0.4f)\n", i, j, v->position.x, v->position.y, v->position.z);
        /* mark this as a dup of the other position */
        v->position_id = u->position_id;
        dup = 1;
        break;
      }
    }
    position = gsm->positions[n];
    copy_vector(&position->local, &v->position);
    position->bone_id = v->bone_id;
    if (! dup) v->position_id = n++;
  }

  /* remove redundant positions */
  for (i = gsm->num_positions - 1; i >= n; i--) {
    TRACE2("freeing redundant new position %u\n", i);
    free(gsm->positions[i]);
    gsm->num_positions--;
  }

  /* make assignments but check for duplicate positions! */
  DEBUG("creating assignments\n");
  /* assume there's one assignment for every bone */
  gsm->assignments = alloc_assignments(model->num_joints);
  if (! gsm->assignments) {
    ERROR("out of memory for assignments");
    return 5;
  }
  gsm->num_assignments = model->num_joints;

  last_bone = BONE_FIRST_LINK;
  assignment_id = -1;
  /* assume vertices were ordered by bone and hence positions now are too */
  for (i = 0; i < gsm->num_positions; i++) {
    position = gsm->positions[i];
    if (position->bone_id != last_bone) {
      /* new assignment */
      if (++assignment_id >= gsm->num_assignments) {
        ERROR("assignment ID %u out of range!\n", assignment_id);
        return 6;
      }
      assignment = gsm->assignments[assignment_id];
      assignment->first_vertex = i;
      assignment->bone_id = v->bone_id;
    }
    assignment->num_vertices++;
    last_bone = v->bone_id;
  }

  /* cleanup unused assignments */
  for (i = gsm->num_assignments - 1; i > 0; i--) {
    assignment = gsm->assignments[i];
    if (assignment->num_vertices) break;
    TRACE2("freeing unused assignment %u\n", i);
    free(assignment);
    gsm->num_assignments--;
  }
  
  return 0;
}

int ms3d_to_gsm_process_triangles(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i, j, num, n, m, vertex_id;
  struct triangle *triangle;
  struct texture_coord *texture_coord, *uv;
  struct position *position;
  struct vertex *vertex, *dup_vertex;
  ms3d_triangle_t *t;
  ms3d_vertex_t *v;
  int dup_vert, dup_uv;
  unsigned short *should_have_used_an_array;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  /* create one vertex for each triangle vertex */
  DEBUG("creating effective vertices\n");
  gsm->num_vertices = model->num_triangles * 3;
  gsm->vertices = alloc_vertices(gsm->num_vertices);
  if (! gsm->vertices) {
    ERROR("out of memory for vertices");
    gsm->num_vertices = 0;
    return 3;
  }

  /* and one texture_coord for each vertex */
  DEBUG("creating texture co-ordinates\n");
  gsm->texture_coords = alloc_texture_coords(gsm->num_vertices);
  if (! gsm->texture_coords) {
    ERROR("out of memory for texture co-ordinates");
    return 4;
  }
  gsm->num_texture_coords = gsm->num_vertices;

  DEBUG("setting triangles\n");
  gsm->triangles = alloc_triangles(model->num_triangles);
  if (! gsm->triangles) {
    ERROR("out of memory for triangles");
    return 5;
  }
  gsm->num_triangles = model->num_triangles;

  n = 0; /* new vertex index */
  m = 0; /* new texture_coord index */
  for (i = 0; i < gsm->num_triangles; i++) {
    triangle = gsm->triangles[i];
    t = model->triangles[i];

    for (num = 0; num < 3; num++) {
      /* reverse order to counter normal flipping */
      if (num == 0) should_have_used_an_array = &triangle->vertex3;
      else if (num == 1) should_have_used_an_array = &triangle->vertex2;
      else if (num == 2) should_have_used_an_array = &triangle->vertex1;

      vertex_id = t->vertex_id[num];
      if (vertex_id >= model->num_vertices) {
        ERROR("vertex %u in triangle %u has ID %u out of range (%u vertices in model)\n", num, i, vertex_id, model->num_vertices);
        return 6;
      }

      /* copy UVs from triangles s and t */
      texture_coord = gsm->texture_coords[m];
      texture_coord->u = t->s[num];
      texture_coord->v = t->t[num];

      /* initially just copy IDs */
      v = model->vertices[vertex_id];
      v->vertex_id = n;

      /* find dup vertices */
      dup_vert = dup_uv = 0;
      for (j = 0; j < n; j++) {
        /* silly variable names */
        dup_vertex = gsm->vertices[j];
        position = gsm->positions[dup_vertex->position_id];
        if (vector_equiv(&v->position, &position->local)) {
          TRACE("triangle %u vertex %u: vertex %u (ms3d_vertex %u) is a dup of vertex %u\n", i, num, n, vertex_id, j);
          uv = gsm->texture_coords[dup_vertex->texture_coord_id];
          v->position_id = dup_vertex->position_id;
          if (texture_coord_equiv(texture_coord, uv)) {
            dup_vert = 1;
            v->vertex_id = *should_have_used_an_array = j;
            break;
          }
          TRACE("but vertex %u's UV is different from vertex %u's - so not a real dup\n", j, n);
        }
      }

      if (! dup_vert) *should_have_used_an_array = n++;
      vertex = gsm->vertices[v->vertex_id];
      vertex->position_id = v->position_id;
      vertex->num_triangles++;

      /* find dup texture co-ordinates */
      for (j = 0; j < m; j++) {
        uv = gsm->texture_coords[j];
        if (texture_coord_equiv(texture_coord, uv)) {
          TRACE("texture co-ordinate %u is a dup of %u\n", m, j);
          dup_uv = 1;
          vertex->texture_coord_id = j;
          break;
        }
      }
      if (! dup_uv) vertex->texture_coord_id = m++;
    }
  }

  /* remove redundant vertices */
  for (i = gsm->num_vertices - 1; i >= n; i--) {
    TRACE2("freeing redundant new vertex %u\n", i);
    free(gsm->vertices[i]);
    gsm->num_vertices--;
  }

  /* remove redundant texture_coords */
  for (i = gsm->num_texture_coords - 1; i >= m; i--) {
    TRACE2("freeing redundant new texture_coord %u\n", i);
    free(gsm->texture_coords[i]);
    gsm->num_texture_coords--;
  }

  return 0;
}

int ms3d_to_gsm_process_groups(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i, n;
  struct triangle_group *triangle_group;
  ms3d_group_t *g;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  DEBUG("making triangle groups\n");
  gsm->triangle_groups = alloc_triangle_groups(model->num_groups);
  if (! gsm->triangle_groups) {
    ERROR("out of memory for triangle groups\n");
    return 3;
  }
  gsm->num_triangle_groups = model->num_groups;

  n = 0;
  for (i = 0; i < gsm->num_triangle_groups; i++) {
    g = model->groups[i];
    /* assume triangles are sorted */
    if (! g->triangle_id[0]) {
      WARN("unassigned material group %u \"%s\"; fix your model!\n", i, g->name);
      continue;
    }
    triangle_group = gsm->triangle_groups[n];
    triangle_group->first_triangle = *g->triangle_id[0];
    triangle_group->num_triangles = g->num_triangles;
    if (g->material_id == -1) {
      ERROR("panic! group '%s' (ID %u) has no material\n", g->name, i + 1);
      return 4;
    }
    triangle_group->material_id = (unsigned short) g->material_id;
    n++;
  }

  /* remove redundant groups */
  for (i = gsm->num_triangle_groups - 1; i > n; i--) {
    TRACE2("removing redundant new triangle_group %u\n", i);
    free(gsm->triangle_groups[i]);
    gsm->num_triangle_groups--;
  }

  return 0;
}

int ms3d_to_gsm_process_materials(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i, texture_id;
  struct material *material;
  struct map *map;
  struct texture *texture;
  ms3d_material_t *m;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  DEBUG("making textures\n");
  gsm->textures = alloc_textures(model->num_materials);
  if (! gsm->textures) {
    ERROR("out of memory for textures\n");
    return 3;
  }
  gsm->num_textures = model->num_materials;

  DEBUG("making materials\n");
  gsm->materials = alloc_materials(model->num_materials);
  if (! gsm->materials) {
    ERROR("out of memory for materials");
    return 4;
  }
  gsm->num_materials = model->num_materials;

  texture_id = 0;
  for (i = 0; i < gsm->num_materials; i++) {
    material = gsm->materials[i];
    m = model->materials[i];
    /* set name */
    material->name = alloc_name(m->name, &material->name_length);
    if (! material->name_length) {
      ERROR("can't set material name\n");
      return 5;
    }
    /* XXX material type isn't really shininess - FIXME */
    material->map_type = MATERIAL_PROPERTY_SHININESS;
    /* ambient */
    material->ambient_red = 255.0 * m->ambient[0] * m->ambient[3];
    material->ambient_green = 255.0 * m->ambient[1] * m->ambient[3];
    material->ambient_blue = 255.0 * m->ambient[2] * m->ambient[3];
    /* diffuse */
    material->diffuse_red = 255.0 * m->diffuse[0] * m->diffuse[3];
    material->diffuse_green = 255.0 * m->diffuse[1] * m->diffuse[3];
    material->diffuse_blue = 255.0 * m->diffuse[2] * m->diffuse[3];
    /* specularity */
    material->specular_red = 255.0 * m->specular[0] * m->specular[3];
    material->specular_green = 255.0 * m->specular[1] * m->specular[3];
    material->specular_blue = 255.0 * m->specular[2] * m->specular[3];
    /* emissive = self-illumination */
    material->self_illumination = m->emissive[3];
    /* set opacity */
    material->transparency = 1.0 - m->opacity;
    /* set shininess */
    material->shininess = m->shininess / 128.0;
    /* default shine strength */
    material->shine_strength = 0.05;

    if (*m->texture) {
      /* make a new texture */
      texture = gsm->textures[texture_id];
      texture->filename = alloc_name(m->texture, &texture->filename_length);
      if (! texture->filename_length) {
        ERROR("can't set texture_name\n");
        return 6;
      }
      set_basename(texture->filename, texture->filename_length, &texture->basename, &texture->basename_length);

      /* make a map of the new texture */
      material->maps = alloc_maps(1);
      if (! material->maps) {
        ERROR("out of memory for material map\n");
        return 7;
      }
      material->num_maps = 1;
      map = material->maps[0];
      map->map_type = MAP_TYPE_DIFFUSE;
      map->texture_id = texture_id;

      texture_id++;
    }
  }

  /* cleanup unused textures */
  DEBUG("cleaning up unused textures\n");
  for (i = gsm->num_textures - 1; i > 0; i--) {
    texture = gsm->textures[i];
    if (texture->filename) break;
    free(texture);
    gsm->num_textures--;
  }

  return 0;
}

int ms3d_to_gsm_process_joints(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i, k;
  struct skeleton *skeleton;
  struct bone *bone;
  struct frame *frame;
  ms3d_joint_t *j;
  int links_to;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  DEBUG("making skeleton\n");
  gsm->skeletons = alloc_skeletons(1);
  if (! gsm->skeletons) {
    ERROR("out of memory for skeleton");
    return 3;
  }
  gsm->num_skeletons = 1;
  skeleton = gsm->skeletons[0];

  DEBUG("making frames\n");
  gsm->frames = alloc_frames(model->num_frames);
  if (! gsm->frames) {
    ERROR("out of memory for frames");
    return 4;
  }
  gsm->num_frames = model->num_frames;

  DEBUG("making bones\n");
  skeleton->bones = alloc_bones(model->num_joints);
  if (! skeleton->bones) {
    ERROR("out of memory for bones");
    return 4;
  }
  skeleton->num_bones = model->num_joints;

  for (i = 0; i < skeleton->num_bones; i++) {
    bone = skeleton->bones[i];
    j = model->joints[i];
    /* set name */
    bone->name = alloc_name(j->name, &bone->name_length);
    if (! bone->name_length) {
      ERROR("can't set bone name\n");
      return 5;
    }
    /* set parent */
    if (*j->parent_name) {
      links_to = get_bone_by_name(gsm, j->parent_name);
      if (links_to < 0) {
        ERROR("panic! can't find parent bone \"%s\"\n", j->parent_name);
        return 6;
      }
      bone->links_to = (unsigned short) links_to;
    }
    else {
      bone->links_to = BONE_FIRST_LINK;
      /* set skeleton name from bone name */
      skeleton->name = alloc_name(bone->name, &skeleton->name_length);
      if (! skeleton->name_length) {
        ERROR("can't set skeleton name\n");
        return 7;
      }
    }
    /* set position */
    copy_vector(&bone->position, &j->position);
    /* set rotation */
    if (euler_to_matrix(&bone->local, &j->rotation)) {
      ERROR("can't convert euler co-ordinates to rotation matrix for bone\n");
      return 8;
    }
    /* XXX: hack, make bone have I rotation */
  {
    matrix_t *m, t;
    if (bone->links_to == BONE_FIRST_LINK) {
      copy_matrix(&t, &bone->local);
      transpose_matrix(&t);
    }
    m = multiply_matrices(&t, &bone->local);
    copy_matrix(&bone->local, m);
    free(m);
  }
    /* XXX: shouldn't be doing this */
    if (matrix_to_quaternion(&bone->q, &bone->local)) {
      ERROR("can't convert euler co-ordinates to quaternion for bone\n");
      return 8;
    }

    /* animation stuff */
    bone->locals = alloc_matrices(model->num_frames);
    if (! bone->locals) {
      ERROR("out of memory for frame rotations");
      return 9;
    }
    /* XXX: shouldn't be using this */
    bone->orientations = alloc_quaternions(model->num_frames);
    if (! bone->orientations) {
      ERROR("out of memory for frame orientations");
      return 9;
    }
    /* rotation keyframes */
    if (j->num_rotation_keyframes) {
      for (k = 0; k < model->num_frames; k++) {
        /* FIXME: use last frame */
        if (k >= j->num_rotation_keyframes) {
          ERROR("rotation keyframe out of range\n");
          return 10;
        }
        if (euler_to_matrix(bone->locals[k], &j->rotation_keyframes[k]->rotation)) {
          ERROR("can't make rotation matrix from euler angles\n");
          return 11;
        }
        /* XXX: shouldn't be using this */
        if (euler_to_quaternion(bone->orientations[k], &j->rotation_keyframes[k]->rotation)) {
          ERROR("can't make rotation matrix from euler angles\n");
          return 11;
        }
      }
    }
    if (bone->links_to == BONE_FIRST_LINK && j->num_position_keyframes) {
      TRACE("setting root bone location in frames\n");
      for (k = 0; k < model->num_frames; k++) {
        /* FIXME: use last frame */
        if (k >= j->num_position_keyframes) {
          ERROR("position keyframe out of range\n");
          return 12;
        }
        frame = gsm->frames[k];
        copy_vector(&frame->root_position, &j->position_keyframes[k]->position);
      }
    }
  }

  return 0;
}

/* make grips at the origin for guns */
int ms3d_to_gsm_create_grips(struct gsm *gsm, ms3d_model_t *model) {
  unsigned short i;
  struct skeleton *skeleton;
  struct bone *bone;
  struct gripper *grip;
  matrix_t *m;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! model) {
    ERROR("no model!\n");
    return 2;
  }

  if (model->num_joints > 2) {
    INFO("biped grippers not supported\n");
    return 0;
  }

  skeleton = gsm->skeletons[0];
  if (! skeleton) {
    ERROR("no skeleton\n");
    return 3;
  }
  bone = skeleton->bones[0];
  if (! bone) {
    ERROR("no bone\n");
    return 4;
  }

  gsm->grips = alloc_grippers(GUN_NUM_GRIPS);
  if (! gsm->grips) {
    ERROR("out of memory for grips");
    return 5;
  }
  gsm->num_grips = GUN_NUM_GRIPS;

  for (i = 0; i < GUN_NUM_GRIPS; i++) {
    grip = gsm->grips[i];
    /* name */
    grip->name = alloc_name(gun_grip_name[i], &grip->name_length);
    if (! grip->name_length) {
      ERROR("out of memory for grip %u name\n", i);
      return 6;
    }
    /* position */
    grip->position.x = 0;
    grip->position.y = 0;
    grip->position.z = 0;
    /* rotation */
    m = multiply_matrices(&transform_ms3d_to_gsm, &bone->local);
    if (! m) {
      ERROR("can't rotate co-ordinate system\n");
      return 7;
    }
    copy_matrix(&grip->rotation, m);
    free(m);
    /* bone */
    grip->bone_id = 0;
  }

  return 0;
}

int write_ms3d_data(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  if (write_ms3d_header(model)) return 3;
  if (write_ms3d_vertices(model)) return 4;
  if (write_ms3d_triangles(model)) return 4;
  if (write_ms3d_groups(model)) return 4;
  if (write_ms3d_materials(model)) return 4;
  if (write_ms3d_animation_stuff(model)) return 4;
  if (write_ms3d_joints(model)) return 4;

  return 0;
}

int write_ms3d_header(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing header\n");
  if (write_arbitrary(model->out_fd, &model->header.magic, sizeof(model->header.magic))) {
    ERROR("can't write MS3D magic\n");
    return 2;
  }
  if (write_int(model->out_fd, &model->header.version)) {
    ERROR("can't write version identifier\n");
    return 3;
  }
  DEBUG("wrote magic \"%s\" version %d\n", model->header.magic, model->header.version);

  return 0;
}

int write_ms3d_vertices(ms3d_model_t *model) {
  unsigned short i;
  ms3d_vertex_t *vertex;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing number of vertices\n");
  if (write_short(model->out_fd, &model->num_vertices)) {
    ERROR("can't write number of vertices\n");
    return 2;
  }

  for (i = 0; i < model->num_vertices; i++) {
    vertex = model->vertices[i];
    /* flags */
    if (write_unsigned_char(model->out_fd, &vertex->flags)) {
      ERROR("can't write vertex %u flags\n", i);
      return 4;
    }
    /* position */
    if (write_vector(model->out_fd, &vertex->position)) {
      ERROR("can't write vertex %u position\n", i);
      return 5;
    }
    /* bone_id */
    if (write_char(model->out_fd, &vertex->bone_id)) {
      ERROR("can't write vertex %u bone ID\n", i);
      return 6;
    }
    /* reference_count */
    if (write_unsigned_char(model->out_fd, &vertex->reference_count)) {
      ERROR("can't write vertex %u reference count\n", i);
      return 7;
    }

    TRACE("wrote vertex %u\n", i);
  }
  DEBUG("wrote %u %s\n", i, (i == 1) ? "vertex" : "vertices");

  return 0;
}

int write_ms3d_triangles(ms3d_model_t *model) {
  unsigned short i;
  ms3d_triangle_t *triangle;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing number of triangles\n");
  if (write_short(model->out_fd, &model->num_triangles)) {
    ERROR("can't write number of triangles\n");
    return 2;
  }

  for (i = 0; i < model->num_triangles; i++) {
    triangle = model->triangles[i];
    /* flags */
    if (write_short(model->out_fd, &triangle->flags)) {
      ERROR("can't write triangle %u flags\n", i);
      return 3;
    }
    /* vertex indices */
    if (write_short(model->out_fd, &triangle->vertex_id[0])) {
      ERROR("can't write triangle %u first vertex\n", i);
      return 4;
    }
    if (write_short(model->out_fd, &triangle->vertex_id[1])) {
      ERROR("can't write triangle %u second vertex\n", i);
      return 5;
    }
    if (write_short(model->out_fd, &triangle->vertex_id[2])) {
      ERROR("can't write triangle %u third vertex\n", i);
      return 6;
    }
    /* normals */
    if (write_vector(model->out_fd, &triangle->normals[0])) {
      ERROR(" can't write triangle %u first normal\n", i);
      return 7;
    }
    if (write_vector(model->out_fd, &triangle->normals[1])) {
      ERROR(" can't write triangle %u second normal\n", i);
      return 8;
    }
    if (write_vector(model->out_fd, &triangle->normals[2])) {
      ERROR(" can't write triangle %u third normal\n", i);
      return 9;
    }
    /* s */
    if (write_float(model->out_fd, &triangle->s[0])) {
      ERROR("can't write triangle %u first s\n", i);
      return 10;
    }
    if (write_float(model->out_fd, &triangle->s[1])) {
      ERROR("can't write triangle %u second s\n", i);
      return 11;
    }
    if (write_float(model->out_fd, &triangle->s[2])) {
      ERROR("can't write triangle %u third s\n", i);
      return 12;
    }
    /* t */
    if (write_float(model->out_fd, &triangle->t[0])) {
      ERROR("can't write triangle %u first t\n", i);
      return 13;
    }
    if (write_float(model->out_fd, &triangle->t[1])) {
      ERROR("can't write triangle %u second t\n", i);
      return 14;
    }
    if (write_float(model->out_fd, &triangle->t[2])) {
      ERROR("can't write triangle %u third t\n", i);
      return 15;
    }
    /* smoothing group */
    if (write_unsigned_char(model->out_fd, &triangle->smoothing_group)) {
      ERROR("can't write triangle %u smoothing group\n", i);
      return 16;
    }
    /* group_id */
    if (write_unsigned_char(model->out_fd, &triangle->group_id)) {
      ERROR("can't write triangle %u group ID\n", i);
      return 17;
    }

    TRACE("wrote triangle %u\n", i);
  }
  DEBUG("wrote %u triangle%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int write_ms3d_groups(ms3d_model_t *model) {
  unsigned short i, j;
  ms3d_group_t *group;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing number of groups\n");
  if (write_short(model->out_fd, &model->num_groups)) {
    ERROR("can't write number of groups\n");
    return 2;
  }

  for (i = 0; i < model->num_groups; i++) {
    group = model->groups[i];
    /* flags */
    if (write_unsigned_char(model->out_fd, &group->flags)) {
      ERROR("can't write group %u flags\n", i);
      return 3;
    }
    /* name */
    if (write_arbitrary(model->out_fd, &group->name, sizeof(group->name))) {
      ERROR("can't write group %u name\n", i);
      return 4;
    }
    /* number of triangles */
    if (write_short(model->out_fd, &group->num_triangles)) {
      ERROR("can't write group %u number of triangles\n", i);
      return 5;
    }
    /* write them out */
    for (j = 0; j < group->num_triangles; j++) {
      if (write_short(model->out_fd, group->triangle_id[j])) {
        ERROR("can't write group %u triangle %u\n", i, j);
        return 6;
      }
      TRACE1("group %u triangle %u: %u\n", i, j, *group->triangle_id[j]);
    }
    /* material_id */
    if (write_char(model->out_fd, &group->material_id)) {
      ERROR("can't write group %u material ID\n", i);
      return 7;
    }

    TRACE("wrote group %u \"%s\"\n", i, group->name);
  }
  DEBUG("wrote %u group%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int write_ms3d_materials(ms3d_model_t *model) {
  unsigned short i;
  ms3d_material_t *material;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing number of materials\n");
  if (write_short(model->out_fd, &model->num_materials)) {
    ERROR("can't write number of materials\n");
    return 2;
  }

  for (i = 0; i < model->num_materials; i++) {
    material = model->materials[i];
    /* name */
    if (write_arbitrary(model->out_fd, &material->name, sizeof(material->name))) {
      ERROR("can't write material %u name\n", i);
      return 3;
    }
    /* ambient */
    if (write_float(model->out_fd, &material->ambient[0])) {
      ERROR("can't write material %u ambient red\n", i);
      return 4;
    }
    if (write_float(model->out_fd, &material->ambient[1])) {
      ERROR("can't write material %u ambient green\n", i);
      return 5;
    }
    if (write_float(model->out_fd, &material->ambient[2])) {
      ERROR("can't write material %u ambient blue\n", i);
      return 6;
    }
    if (write_float(model->out_fd, &material->ambient[3])) {
      ERROR("can't write material %u ambient alpha\n", i);
      return 7;
    }
    /* diffuse */
    if (write_float(model->out_fd, &material->diffuse[0])) {
      ERROR("can't write material %u diffuse red\n", i);
      return 8;
    }
    if (write_float(model->out_fd, &material->diffuse[1])) {
      ERROR("can't write material %u diffuse green\n", i);
      return 9;
    }
    if (write_float(model->out_fd, &material->diffuse[2])) {
      ERROR("can't write material %u diffuse blue\n", i);
      return 10;
    }
    if (write_float(model->out_fd, &material->diffuse[3])) {
      ERROR("can't write material %u diffuse alpha\n", i);
      return 11;
    }
    /* specular */
    if (write_float(model->out_fd, &material->specular[0])) {
      ERROR("can't write material %u specular red\n", i);
      return 12;
    }
    if (write_float(model->out_fd, &material->specular[1])) {
      ERROR("can't write material %u specular green\n", i);
      return 13;
    }
    if (write_float(model->out_fd, &material->specular[2])) {
      ERROR("can't write material %u specular blue\n", i);
      return 14;
    }
    if (write_float(model->out_fd, &material->specular[3])) {
      ERROR("can't write material %u specular alpha\n", i);
      return 15;
    }
    /* emissive */
    if (write_float(model->out_fd, &material->emissive[0])) {
      ERROR("can't write material %u emissive red\n", i);
      return 16;
    }
    if (write_float(model->out_fd, &material->emissive[1])) {
      ERROR("can't write material %u emissive green\n", i);
      return 17;
    }
    if (write_float(model->out_fd, &material->emissive[2])) {
      ERROR("can't write material %u emissive blue\n", i);
      return 18;
    }
    if (write_float(model->out_fd, &material->emissive[3])) {
      ERROR("can't write material %u emissive alpha\n", i);
      return 19;
    }
    /* shininess */
    if (write_float(model->out_fd, &material->shininess)) {
      ERROR("can't write material %u shininess\n", i);
      return 20;
    }
    /* opacity */
    if (write_float(model->out_fd, &material->opacity)) {
      ERROR("can't write material %u opacity\n", i);
      return 21;
    }
    /* mode */
    if (write_char(model->out_fd, &material->mode)) {
      ERROR("can't write material %u mode\n", i);
      return 22;
    }
    /* bitmap */
    if (write_arbitrary(model->out_fd, &material->texture, sizeof(material->texture))) {
      ERROR("can't write material %u texture name\n", i);
      return 23;
    }
    /* alpha map */
    if (write_arbitrary(model->out_fd, &material->alphamap, sizeof(material->alphamap))) {
      ERROR("can't write material %u alpha map name\n", i);
      return 24;
    }

    TRACE("wrote material %u \"%s\"\n", i, material->texture);
  }
  DEBUG("wrote %u material%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int write_ms3d_animation_stuff(ms3d_model_t *model) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  /* framerate */
  if (write_float(model->out_fd, &model->framerate)) {
    ERROR("can't write framerate\n");
    return 2;
  }
  DEBUG("wrote framerate: %f fps\n", model->framerate);
  /* current time */
  if (write_float(model->out_fd, &model->current_time)) {
    ERROR("can't write time\n");
    return 3;
  }
  DEBUG("wrote time: %f\n", model->current_time);
  /* frames */
  if (write_int(model->out_fd, &model->num_frames)) {
    ERROR("can't write number of frames\n");
    return 4;
  }
  DEBUG("wrote frames: %d\n", model->num_frames);

  return 0;
}

int write_ms3d_joints(ms3d_model_t *model) {
  unsigned short i, j;
  ms3d_joint_t *joint;

  if (! model) {
    ERROR("no model!\n");
    return 1;
  }

  DEBUG("writing number of joints\n");
  if (write_short(model->out_fd, &model->num_joints)) {
    ERROR("can't write number of joints\n");
    return 2;
  }

  for (i = 0; i < model->num_joints; i++) {
    joint = model->joints[i];
    /* flags */
    if (write_unsigned_char(model->out_fd, &joint->flags)) {
      ERROR("can't write joint %u flags\n", i);
      return 3;
    }
    /* name */
    if (write_arbitrary(model->out_fd, &joint->name, sizeof(joint->name))) {
      ERROR("can't write joint %u name\n", i);
      return 4;
    }
    /* parent's name */
    if (write_arbitrary(model->out_fd, &joint->parent_name, sizeof(joint->parent_name))) {
      ERROR("can't write joint %u parent name\n", i);
      return 5;
    }
    /* rotation */
    if (write_euler(model->out_fd, &joint->rotation)) {
      ERROR("can't write joint %u rotation\n", i);
      return 6;
    }
    /* position */
    if (write_vector(model->out_fd, &joint->position)) {
      ERROR("can't write joint %u position\n", i);
      return 7;
    }
    /* rotation keyframes */
    if (write_short(model->out_fd, &joint->num_rotation_keyframes)) {
      ERROR("can't write joint %u number of rotation keyframes\n", i);
      return 8;
    }
    /* position keyframes */
    if (write_short(model->out_fd, &joint->num_position_keyframes)) {
      ERROR("can't write joint %u number of position keyframes\n", i);
      return 9;
    }
    /* write them out */
    for (j = 0; j < joint->num_rotation_keyframes; j++) {
      if (write_float(model->out_fd, &joint->rotation_keyframes[j]->time)) {
        ERROR("can't write joint %u rotation keyframe %u time\n", i, j);
        return 10;
      }
      if (write_euler(model->out_fd, &joint->rotation_keyframes[j]->rotation)) {
        ERROR("can't write joint %u rotation keyframe %u rotation\n", i, j);
        return 11;
      }
      TRACE1("wrote joint %u rotation keyframe %u\n", i, j);
    }
    /* write them out */
    for (j = 0; j < joint->num_position_keyframes; j++) {
      if (write_float(model->out_fd, &joint->position_keyframes[j]->time)) {
        ERROR("can't write joint %u position keyframe %u time\n", i, j);
        return 12;
      }
      if (write_vector(model->out_fd, &joint->position_keyframes[j]->position)) {
        ERROR("can't write joint %u position keyframe %u position\n", i, j);
        return 13;
      }
      TRACE1("wrote joint %u position keyframe %u\n", i, j);
    }

    TRACE("wrote joint %u \"%s\" linked to \"%s\"\n", i, joint->name, joint->parent_name);
  }
  DEBUG("wrote %u joint%s\n", i, (i == 1) ? "" : "s");

  return 0;
}

int gsm_to_ms3d(ms3d_model_t *model, struct gsm *gsm) {
  if (! model) {
    ERROR("no model!\n");
    return 1;
  }
  if (! gsm) {
    ERROR("no GSM!\n");
    return 2;
  }

  ERROR("you can't do this!\n");
  return 666;

  SAY("converting GSkin data to Milkshape 3D format\n");
  if (gsm_to_ms3d_create_vertices(model, gsm)) return 3;
  if (gsm_to_ms3d_create_triangles(model, gsm)) return 4;
  if (gsm_to_ms3d_create_groups(model, gsm)) return 5;
  if (gsm_to_ms3d_create_materials(model, gsm)) return 6;
  if (gsm_to_ms3d_create_joints(model, gsm)) return 7;
  if (gsm_to_ms3d_create_animation_stuff(model, gsm)) return 8;

  return 0;
}

int gsm_to_ms3d_create_vertices(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!! !!!!!!!! !!!!!!!!!!!!!!!;
  !!!!!!!!!!!!! !!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !
  
  !!!!!!!!!!!!!!! !!!!!!!!!!!!;
  !!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!!!!!!!!! !
    !!!!!!!!!! !! !!!!!! !!! !!!!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!;

  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!!!;
    !!!!!!!! ! !!!!!!!!!!!!!!!!!;
    /* flags */
    !!!!!!!! ! !;
    /* (should be world?) position */
    !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!!!!!!!;
    /* bone */
    !!!!!!!!!! ! !!!!!!!!!!!!!!!!!;
    /* triangle reference count */
    !!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!;
    !! !!!!!!!!!!!!!!!! ! !! !
      !!!!!!!!!!!! !!! !!!!!!!!! !!!!! !!! !!!!!!!! !!!!!! !!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!!!! ! !!!!!!!!! !!!!!! !!!!!!!!!!!!!!!;
  !

  !!!!!! !;
#endif
}

int gsm_to_ms3d_create_triangles(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !;
  !!! !!!!!!!!;
  !!!!!! !!!!!! !!!!!!!;
  !!!!!! !!!!!!!!!!!!! !!!!!!!!!!!!!!;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!!!!! !!!!!!!! !!! !!! !!;
  !!!!!!!!!!!!!!! !!;
  !!!!!!!! !!! !!! !!!!!!!!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !
  
  !!!!!!!!!!!!!!! !!!!!!!!!!!!!;
  !!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!!!!!!!!!! !
    !!!!!!!!!! !! !!!!!! !!! !!!!!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!;

  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!!!!;
    !!!!!!!! ! !!!!!!!!!!!!!!!!!;
    /* flags */
    !!!!!!!! ! !;
    /* vertices and UVs */
    /* first */
    !!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!! !
      !!!!!!!!!!! !!!!!! !!!!!! !!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!!!!!!!!! !
      !!!!!!!!!!! !!!!!!! !!!!!!!!!!! !!!!!! !!!!!!!!!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!;
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    /* second */
    !!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!! !
      !!!!!!!!!!! !!!!!! !!!!!! !!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!!!!!!!!! !
      !!!!!!!!!!! !!!!!!! !!!!!!!!!!! !!!!!! !!!!!!!!!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!;
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    /* third */
    !!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!! !
      !!!!!!!!!!! !!!!!! !!!!!! !!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!!!!!!!!! !
      !!!!!!!!!!! !!!!!!! !!!!!!!!!!! !!!!!! !!!!!!!!!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!;
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!!;
    /* normals */
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!! !!;
    !!!!!!!!!!!!!!!!!!!;
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!! !!;
    !!!!!!!!!!!!!!!!!!!;
    !!!!!!! ! !!!!!!!!!!!!!!!! !!;
    !!!!!!!;
    !!!!!!!;
    !! !! !!!!!!!! !
      !!!!!!!!!!!! !!!! !!!!! !!!!!!! !! !!!! !!!!!!! !!! !!!!!!!! !!!!!! !!;
      !!!!!! !!;
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!;
    !!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!;
    !!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!;
    /* smoothing group */
    !!!!!!!!!!!!!!!!!! ! !;
    /* group */
    !!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!;
    !! !!!!!!!!! ! !! !
      !!!!!!!!!!!! !!! !!!!! !! !!! !!!!!!!! !!!!!! !!;
      !!!!!! !!;
    !
    !! !!!!!!!!! ! !!!!! !
      !!!!!!!!!!!! !! !!! !! !!!!!! !!!! !!!!!! !!!! !!!!!!!! !!! !!!!!!!!!!;
      !!!!!! !!;
    !
    !!!!!!!!!!! ! !!!!!!!!! !!!!! !!!!!!!!;
  !

  !!!!!! !;
#endif
}

int gsm_to_ms3d_create_groups(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !! !;
  !!!! !!!!!!!!!!!!!!!!!!!!!;
  !!!!!! !!!!!!!!!!!!!! !!!!!!!!!!!!!!!;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!!!!!!!!!!! !!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !

  !!!!!!!!!!!!!!! !!!!!!!!!!;
  !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!!!!!!! !
    !!!!!!!!!! !! !!!!!! !!! !!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!;

  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!;
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!;
    !!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!!!! !
      !!!!!!!!! !!!!!!!! !!!!!! !!!!!!!!!!!!!!!!!!!!!!!!!!!!;
      !!!!!! !;
    !
    /* flags */
    !!!!!!!! ! !;
    /* name */
    !! !!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!! !!!!!!!!!!!! ! !! !
      !!!!!!!!!!!! !!!!!!!! !!!!!!!! !!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!!! !!!!! !!!!!!!!!!!!!!!!!;
    !!!!!!!!!!!!!! !!!!! !!!!!!!!!! !!!!!!!!;
    /* triangles */
    !!!!!!!!!!!!!! ! !!!!!!!!! !!!!! !!! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!! !!!!!!!!!!!!!!! !!!!! !!! !!!!! !!!!!!! !!!!!!!!!!!!;
    !! !! !!!!!!!!!!!!!!! !
      !!!!!!!!!! !! !!!!!! !!! !!!!! !!!!!! !!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !!! !! ! !; ! ! !!!!!!!!!!!!!!!!; !!!! !
      !!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ! !;
    !
    /* material ID */
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !

  !!!!!! !;
#endif
}

int gsm_to_ms3d_create_materials(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!!!!! !!! !!!!;
  !!!!!! !!!!!!! !!!!!!!!;
  !!!! !!!!!!!!!!!!!!!!!!!!!;
  !!!! !!!!!!!!!!!!!!!!!!!!!!!!;
  !!!!!!!!!!!!!!! !!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !

  !!!!!!!!!!!!!!! !!!!!!!!!!!!!;
  !!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!!!!!!!!!! !
    !!!!!!!!!! !! !!!!!! !!! !!!!!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!;

  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!!!!;
    !!!!!!!! ! !!!!!!!!!!!!!!!!!;
    /* name */
    !! !!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!! !!!!!!!!!!!! ! !! !
      !!!!!!!!!!!! !!!!!!!! !!!!!!!! !!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!!! !!!!! !!!!!!!!!!!!;
    /* ambient */
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!;
    /* diffuse */
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!;
    /* specular */
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!! ! !!!!!;
    !!!!!!!!!!!!! ! !!!;
    /* emissive */
    !!!!!!!!!!!!!! ! !;
    !!!!!!!!!!!!!! ! !;
    !!!!!!!!!!!!!! ! !;
    !!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!;
    /* shininess */
    !!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!! ! !!!!!;
    /* opacity */
    !!!!!!!!!! ! !!! ! !!!!!!!!!!!!!!!!!!!!!!;
    /* wtf is mode anyway? */
    !!!!!!! ! !;
    /* bitmap */
    !! !!!!!!!!!!!!!!!!!!!! !
      !!! ! !!!!!!!!!!!!!!!!!;
      !! !! !!!! !
        !!!!!!!!!!! !!! !!! !!!!!!!! !!!!!!!!!! !!!!!!!!!!!!!!!;
        !!!!!! !;
      !
      !!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
      !! !! !!!!!!!! !
        !!!!!!!!!!! !!!!!!! !!!!!! !!!!!!!!!!!!!!!!;
        !!!!!! !;
      !
      !! !!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!!!!! !!!!!!!!!!!! ! !! !
        !!!!!!!!!!!! !!!!!!!! !!!!!!! !!!!!!!!;
        !!!!!! !;
      !
      !!!!!!!!!!!!!!!!!!!! !!!!! !!!!!!!!!!!!;
    !
    /* else; */
    /* alphamap is a no-op 'cos alloc_ms3d_material() memset() it to 0 */
  !

  !!!!!! !;
#endif
}

int gsm_to_ms3d_create_joints(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!!!!! !!!! !!!!!! !!!!!!!;
  !!!! !!!!!!!!!!!!!!!!!!!!!;
  !!!!!!!!!!!! !!;
  !!!!!!!! !!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!! !! ! !!!!!!!!!!!!!!!!!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!! ! !!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!! !
    !!!!!!!!! !!!!!!!!!!!!!;
    !!!!!! !;
  !

  !!!!!!!!!!!!!!! !!!!!!!!!!;
  !!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!!!!!!! !
    !!!!!!!!!! !! !!!!!! !!! !!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!;

  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!;
    !!!! ! !!!!!!!!!!!!!!!!!!;
    /* flags */
    !!!!!!!! ! !;
    /* name */
    !! !!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!! !!!!!!!!!!!! ! !! !
      !!!!!!!!!!!! !!!!!!!! !!!! !!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!!! !!!!! !!!!!!!!!!!!;
    /* parent name */
    !! !!!!!!!!!!!!!!! !! !!!!!!!!!!!!!!!! !
      !! !!!!!!!!!!!!!!! !! !!!!!!!!!!!!!!!!!!!! !
        !!!!!!!!!!!!! !!!! !! !! !!! !! !!!!!!!!! !!!!!!!!!!!!!!!;
        !!!!!! !;
      !
      !!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
      !! !! !!!!!!! !
        !!!!!!!!!!! !!!!!! !!!!!!!!;
        !!!!!! !;
      !
      !! !!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!! !!!!!!!!!!!! ! !! !
        !!!!!!!!!!!! !!!!!!!! !!!!!! !!!! !!!!!!!!;
        !!!!!! !;
      !
      !!!!!!!!!!!!!!!!!!!!!!!! !!!!! !!!!!!!!!!!!;
    !
    /* if it's the root bone it's a no-op as parent name is already 0 */
    /* rotation */
    DEBUG("boo\n");
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    DEBUG("goose\n");
    ! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!;
    !! !! !! !
      !!!!!!!!!!!! !!!!!!!!! !!!!!!!!!!! !!!!!!!!!!;
      !!!!!! !;
    !
    !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!! !
      !!!!!!!!!!!! !!!!!!! !!!!! !!!!!!!! !!!!!! !! !!!!! !!!!!!!!!!;
      !!!!!! !!;
    !
    !!!!!!!;
    /* (local) position */
    !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!!!;
  !

  !!!!!! !;
#endif
}

int gsm_to_ms3d_create_animation_stuff(ms3d_model_t *model, struct gsm *gsm) {
#if 0
  !!!!!!!! !!!!! !! !;
  !!!!!! !!!!!!!! !!!!!!!!!;
  !!!!!! !!!! !!!!!! !!!!!!!;
  !!!! !!!!!!!!!!!!!!!!!!!!!;
  !!!!!!!!!!!! !!;
  !!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!!!!!;

  !! !! !!!!!! !
    !!!!!!!!! !!!!!!!!!!;
    !!!!!! !;
  !
  !! !! !!! !! ! !!!!!!!!!!!!!!!!!!! !
    !!!!!!!!! !!!!!!!!;
    !!!!!! !;
  !
  !!!!!!!! ! !!!!!!!!!!!!!!!!!;
  !! !! !!!!!!!!! !
    !!!!!!!!! !!!!!!!!!!!!!;
    !!!!!! !;
  !

  !!!!!!!!!!!!!!! !!!!!!!!!!;
  !!!!!!!!!!!!!!!! ! !!!!;
  !!!!!!!!!!!!!!!!!!! ! !;

  !! !! !!!!!!!!!!!!!!!!!! !
    !!!!!!!!!!!! !!!! !! !!! !! !!!!!!!!! ! !!!!!!! !!!! !! !!!!!!;
    !!!!!!!!!!!!!!!!! ! !;
    !!!!!! !;
  !
  !!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!;

  /* create keyframes for all bones */
  !!! !! ! !; ! ! !!!!!!!!!!!!!!!!!; !!!! !
    ! ! !!!!!!!!!!!!!!!!;
    !!!! ! !!!!!!!!!!!!!!!!!!;
    /* allocate rotation keyframes */
    !!!!!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
    !! !! !!!!!!!!!!!!!!!!!!!!!! !
      !!!!!!!!!! !! !!!!!! !!! !!!!!!!! !!!!!!!!!!!;
      !!!!!! !;
    !
    !!!!!!!!!!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!;

    /* FIXME */

    !!! !! ! !; ! ! !!!!!!!!!!!!!!!; !!!! !
      !!!!!!!!!!!!!!!!! ! !!!!!!!!!!!!!!!!!!!!!!!!;
      !!!!!!!!!!!!!!!!!!!!!!! ! !!!!!!! !;
      !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!!!!!!!!!!! !
        !!!!!!!!!!!! !!!!!!! !!!!! !!!!!!!! !!!!!! !!! !!!! !! !! !!!!! !!!!!! !! !!;
        !!!!!! !;
      !
    !
  !

  !!!!!! !;
#endif
}
