#include "gsmhacking.h"

int optimise_materials(struct gsm *gsm) {
  unsigned short i, j, k, count;

  if (! gsm) {
    ERROR("no gsm");
    return -1;
  }

  count = 0;
  DEBUG("removing redundant materials\n");
  for (i = 0; i < gsm->num_materials; i++) {
    if (get_num_triangle_groups(gsm, i)) continue;
    DEBUG("material %02u %s is not mapped to any triangle groups\n", i, gsm->materials[i]->name);
    free(gsm->materials[i]->name);
    free(gsm->materials[i]);

    TRACE1("shifting materials from %u\n", i);
    for (j = i; j < gsm->num_materials - 1; j++) {
      if (j > i) {
        if (gsm->materials[j + 1]->number != j + 1) {
          WARN("material %02u has number %02u (not %02u)\n", j + 1, gsm->materials[j + 1]->number, j + 1);
          WARN("GSM is probably going to be destroyed\n");
        }
      }
      gsm->materials[j] = gsm->materials[j + 1];
      gsm->materials[j]->number = j;
    }

    /* assumption */
    for (k = 0; k < gsm->num_triangle_groups; k++) {
      if (gsm->triangle_groups[k]->material_id > i) gsm->triangle_groups[k]->material_id--;
    }

    gsm->num_materials--;
    count++;
    i--;
  }

  if (count == 1) {
    SAY("removed one redundant material\n");
  }
  else if (count) {
    SAY("removed %u redundant materials\n", count);
  }
  else {
    ERROR("materials are optimised\n");
  }

  return 0;
}

int optimise_textures(struct gsm *gsm) {
  unsigned short i, j, k, l, count;

  if (! gsm) {
    ERROR("no gsm");
    return -1;
  }

  count = 0;
  DEBUG("removing redundant textures\n");
  for (i = 0; i < gsm->num_textures; i++) {
    if (get_num_maps(gsm, i)) continue;
    DEBUG("texture %02u %s is not mapped to any materials\n", i, gsm->textures[i]->basename);
    free(gsm->textures[i]->filename);
    free(gsm->textures[i]);

    TRACE1("shifting textures from %u\n", i);
    for (j = i; j < gsm->num_textures - 1; j++) {
      gsm->textures[j] = gsm->textures[j + 1];
    }

    /* assumption */
    for (k = 0; k < gsm->num_materials; k++) {
      for (l = 0; l < gsm->materials[k]->num_maps; l++) {
        if (gsm->materials[k]->maps[l]->texture_id > i) gsm->materials[k]->maps[l]->texture_id--;
      }
    }

    gsm->num_textures--;
    count++;
    i--;
  }

  if (count == 1) {
    SAY("removed one redundant texture\n");
  }
  else if (count) {
    SAY("removed %u redundant textures\n", count);
  }
  else {
    ERROR("textures are optimised\n");
  }

  /*
    it's all very well optimising out duplicate texture names for multiple 
    materials but then it's a pain in the ass if you really want to use two 
    textures later
  */
  INFO("not removing duplicate texture filenames\n");

  return 0;
}

int optimise(struct gsm *gsm) {
  if (gsm->type & GSM_TYPE_ANIMATION) return optimise_gsa(gsm);
  else return optimise_gsm(gsm);
}

int optimise_gsm(struct gsm *gsm) {
  if (! gsm || ! gsm->basename) {
    ERROR("no gsm");
    return -1;
  }

  if (! gsm->need_optimise) {
    SAY("%s doesn't need optimising\n", gsm->basename);
    return 0;
  }

  if (optimise_materials(gsm)) return 1;
  if (optimise_textures(gsm)) return 2;
  if (gsm->type == GSM_TYPE_PIS) if (make_rifle(gsm)) return 3;

  return 0;
}

/* animations made with version 4.1 of the exporter don't grok events */
int optimise_gsa(struct gsm *gsa) {
  if (! gsa || ! gsa->basename) {
    ERROR("no gsa");
    return -1;
  }

  if (! (gsa->version >> 8 == 4 && gsa->version % 16 < 2)) {
    SAY("%s is version %u.%u, doesn't need optimising\n", gsa->version >> 8, gsa->version % 16, gsa->basename);
    return 0;
  }

  INFO("marking animation as version 4.2\n");
  gsa->version = 1026;
  TRACE("no need to initialise event structs, alloc_bone() does that for us\n");

  return 0;
}
