#include "gsmhacking.h"

struct position **alloc_positions(unsigned short n) {
  return (struct position **) alloc_lots(n, "position", sizeof(struct position *), (void *(*)()) alloc_position);
}

struct position **alloc_more_positions(unsigned short n, unsigned short num_positions, struct position **positions) {
  return (struct position **) alloc_more(n, "position", sizeof(struct position *), (void *(*)()) alloc_position, num_positions, (void **) positions, sizeof(struct position));
}

struct position *alloc_position() {
  return (struct position *) alloc_one("position", sizeof(struct position));
}

void cleanup_positions(unsigned short num_positions, struct position **positions) {
  free_structs("position", num_positions, (void **) positions);
}

/* find out how many positions there are and what their offsets are */
int find_positions(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct position *position;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  if (gsm->need_assignments_hack) {
    if (read_short(gsm->fd, &gsm->first_blended_position)) {
      ERROR("can't read first blended position");
      return 2;
    }
    if (read_int(gsm->fd, &gsm->number_blended_positions)) {
      ERROR("can't read number of blended positions (int)");
      return 3;
    }
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think positions start at 0x%04x\n", pos);

  /* how many positions are there? */
  if (read_short(gsm->fd, &gsm->num_positions)) {
    ERROR("can't read number of positions");
    return 4;
  }
  if (gsm->number_blended_positions) {
    if (gsm->num_positions != gsm->first_blended_position) {
      ERROR("number of positions %u doesn't match first blended position ID %u\n", gsm->num_positions, gsm->first_blended_position);
      return 5;
    }
  }
  if (! gsm->num_positions) {
    SAY("%s has no positions\n", gsm->filename);
    return 5;
  }
  DEBUG("positions: %u\n", gsm->num_positions);

  /* should be true because we SHOULD have parsed them out already */
  if (gsm->num_assignments) {
    if (gsm->num_positions != gsm->assignments[gsm->num_assignments - 1]->first_vertex + gsm->assignments[gsm->num_assignments - 1]->num_vertices) {
      ERROR("number of positions (%u) doesn't match number of assignments (%u)\n", gsm->num_positions, gsm->assignments[gsm->num_assignments - 1]->first_vertex + gsm->assignments[gsm->num_assignments - 1]->num_vertices);
    }
  }
  else WARN("gsm->num_assignments undefined\n");

  /* allocate a big enough struct */
  gsm->positions = alloc_positions(gsm->num_positions);
  if (! gsm->positions) return 4;

  /* fill in the positions */
  for (i = 0; i < gsm->num_positions; i++) {
    position = gsm->positions[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    position->offset = pos;
    if (parse_position(gsm, i, position, gsm->num_positions)) {
      ERROR("expected position id %u but got null\n", i);
      return 6;
    }
    if (calculate_position_matrices(gsm, i, position)) {
      ERROR("can't calculate position %u's world\n", i);
      return 7;
    }
  }
  if (assign_bones(gsm)) {
    ERROR("can't set bone ID for each position\n");
    return 8;
  }

  return 0;
}

int parse_position(struct gsm *gsm, unsigned short num, struct position *position, unsigned short total) {
  if (read_vector(gsm->fd, &position->local)) {
    ERROR("read failed parsing position co-ordinates\n");
    return 1;
  }

  TRACE("position %u at 0x%04x: (% 0.4f, % 0.4f, % 0.4f)\n", num, position->offset, position->local.x, position->local.y, position->local.z);

  return 0;
}

int write_positions(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  if (gsm->need_assignments_hack) {
    if (write_short(gsm->out_fd, &gsm->first_blended_position)) {
      ERROR("can't write first blended position");
      return 2;
    }
    if (write_int(gsm->out_fd, &gsm->number_blended_positions)) {
      ERROR("can't write number of blended position (int)s");
      return 3;
    }
  }

  DEBUG("writing positions: %u\n", gsm->num_positions);
  if (write_short(gsm->out_fd, &gsm->num_positions)) {
    ERROR("can't write number of positions");
    return 4;
  }

  for (i = 0; i < gsm->num_positions; i++) {
    TRACE("writing position %u\n", i);
    if (write_vector(gsm->out_fd, &gsm->positions[i]->local)) {
      ERROR("can't write position %u co-ordinates", i);
      return 5;
    }
  }

  return 0;
}

/* no matrices really, just keeping with the naming convention */
int calculate_position_matrices(struct gsm *gsm, unsigned short position_id, struct position *position) {
  struct bone *bone, *root;
  vector_t *v;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! position) {
    ERROR("no position\n");
    return 2;
  }

  if (position->bone_id >= gsm->skeletons[0]->num_bones) {
    ERROR("position %u bone ID %u out of range\n", position_id, position->bone_id);
    return 3;
  }

  /* this code tested to give same answer as PortaLib implementation */
  bone = gsm->skeletons[0]->bones[position->bone_id];
  root = gsm->skeletons[0]->bones[0];
  v = multiply_vector_by_matrix(&position->local, &bone->local);
  if (! v) {
    ERROR("can't find position %u's position in world co-ordinates\n", position_id);
    return 4;
  }
  v->x += bone->world[TX];
  v->y += bone->world[TY];
  v->z += bone->world[TZ];
#if 0
  v->x += root->world[TX];
  v->y += root->world[TY];
  v->z += root->world[TZ];
#endif
  copy_vector(&position->world, v);
  free(v);

  return 0;
}

int describe_positions(struct gsm *gsm) {
  unsigned short i;
  struct position *position;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Vertices: %u\n", gsm->num_positions);
  if (gsm_verbose > 1) {
    for (i = 0; i < gsm->num_positions; i++) {
      position = gsm->positions[i];
      SAY("  Vertex %04u at 0x%04x: (% 0.4f, % 0.4f, % 0.4f)\n", i, position->offset, position->world.x, position->world.y, position->world.z);
      if (gsm_verbose > 2) {
        SAY("    Local co-ordinates: (% 0.4f, % 0.4f, % 0.4f)\n", i, position->offset, position->local.x, position->local.y, position->local.z);
      }
    }
  }

  return 0;
}
